(* $Id: models3.ml 189 2008-02-26 11:51:58Z reuter $ *)
(* Copyright (C) 2000-2004 by Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
   O'Mega is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by 
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   O'Mega is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  *)  

let rcs_file = RCS.parse "Models3" ["extended SUSY models"]
    { RCS.revision = "$Revision: 189 $";
      RCS.date = "$Date: 2008-02-26 12:51:58 +0100 (Tue, 26 Feb 2008) $";
      RCS.author = "$Author: reuter $";
      RCS.source
        = "$URL: http://theorie.physik.uni-wuerzburg.de/svn/ohl/collabs/event-generators/whizard/branches/1.xx/omega-src/bundle/src/models3.ml $" }

(* \thocwmodulesection{Next-to-Minimal Supersymmetric Standard Model} *)

module type extMSSM_flags =
  sig
    val ckm_present       : bool
    val nmssm             : bool
    val exotics           : bool
  end

module NMSSM : extMSSM_flags =
  struct 
    let ckm_present       = false
    let nmssm             = true
    let exotics           = false
  end

module E6SSM : extMSSM_flags =
  struct 
    let ckm_present       = false
    let nmssm             = true
    let exotics           = true
  end


module ExtMSSM (Flags : extMSSM_flags) = 
  struct
    let rcs = RCS.rename rcs_file "Models.NMSSM"
        [ "NMSSM" ]

    open Coupling

    let default_width = ref Timelike
    let use_fudged_width = ref false

    let options = Options.create
      [ "constant_width", Arg.Unit (fun () -> default_width := Constant),
        "use constant width (also in t-channel)";
        "fudged_width", Arg.Set use_fudged_width,
        "use fudge factor for charge particle width";
        "custom_width", Arg.String (fun f -> default_width := Custom f),
        "use custom width";
        "cancel_widths", Arg.Unit (fun () -> default_width := Vanishing),
        "use vanishing width"]

    type gen = 
      | G of int | GG of gen*gen

    let rec string_of_gen = function
      | G n when n > 0  -> string_of_int n
      | G n -> string_of_int (abs n) ^ "c" 
      | GG (g1,g2) -> string_of_gen g1 ^ "_" ^ string_of_gen g2

(* With this we distinguish the flavour. *)

    type sff = 
      | SL | SN | SU | SD

    let string_of_sff = function
      | SL -> "sl" | SN -> "sn" | SU -> "su" | SD -> "sd"         

(* With this we distinguish the mass eigenstates. At the moment we have to cheat 
   a little bit for the sneutrinos. Because we are dealing with massless 
   neutrinos there is only one sort of sneutrino. *)

    type sfm =
      | M1 | M2

    let string_of_sfm = function 
      | M1 -> "1" | M2 -> "2"

(* We also introduce special types for the charginos and neutralinos. *)

    type char = 
      | C1 | C2 | C1c | C2c

    type neu =
      | N1 | N2 | N3 | N4 | N5

    let int_of_char = function
      | C1 -> 1 | C2 -> 2 | C1c -> -1 | C2c -> -2

    let string_of_char = function
      | C1 -> "1" | C2 -> "2" | C1c -> "-1" | C2c -> "-2"

    let conj_char = function
      | C1 -> C1c | C2 -> C2c | C1c -> C1 | C2c -> C2

    let string_of_neu = function
      | N1 -> "1" | N2 -> "2" | N3 -> "3" | N4 -> "4" | N5 -> "n5"

(* Also we need types to distinguish the Higgs bosons. We follow the 
   conventions of Kuroda, which means
   \begin{align}
   \label{eq:higgs3}
   H_1 &= 
      \begin{pmatrix} 
          \frac{1}{\sqrt{2}} 
          \bigl( 
          v_1 + H^0 \cos\alpha - h^0
          \sin\alpha + \ii A^0 \sin\beta - \ii \phi^0 \cos\beta 
          \bigr) \\ 
          H^- \sin\beta - \phi^- \cos\beta 
       \end{pmatrix},
       \\ & \notag \\
   H_2 & = 
      \begin{pmatrix} 
          H^+ \cos\beta + \phi^+ \sin\beta \\
          \frac{1}{\sqrt{2}}
          \bigl( 
           v_2 + H^0 \sin\alpha + h^0 \cos\alpha + \ii A^0 \cos\beta + 
           \ii \phi^0 \sin\beta 
           \bigr)
      \end{pmatrix} 
      \label{eq:higgs4}
   \end{align}
   This is a different sign convention compared to, e.g., 
   Weinberg's volume iii. We will refer to it as [GS+]. 
*)

    type higgs =
      | H1   (* the light scalar Higgs *)
      | H2   (* the heavy scalar Higgs *)
      | H3   (* the pseudoscalar Higgs *)
      | H4   (* the charged Higgs *)
      | H5   (* the neutral Goldstone boson *)
      | H6   (* the charged Goldstone boson *)
      | DH of higgs*higgs

    let rec string_of_higgs = function
      | H1 -> "h1" | H2 -> "h2" | H3 -> "h3" | H4 -> "h4" 
      | H5 -> "p1" | H6 -> "p2" 
      | DH (h1,h2) -> string_of_higgs h1 ^ string_of_higgs h2

    type flavor =
      | L of int | N of int
      | U of int | D of int
      | Sup of sfm*int | Sdown of sfm*int 
      | Ga | Wp | Wm | Z | Gl 
      | Slepton of sfm*int | Sneutrino of int 
      | Neutralino of neu | Chargino of char 
      | Gluino
      | H_Heavy | H_Light | Hp | Hm | A 
      | S0 | P0
            
    type gauge = unit

    let gauge_symbol () =
      failwith "Models.ExtMSSM.gauge_symbol: internal error"       

    let colsymm = function
      | _ -> (0,false),(0,false)

(* At this point we will forget graviton and -tino. *) 

    let family g = [ L g; N g; Slepton (M1,g); 
                     Slepton (M2,g); Sneutrino g;
                     U g; D g; Sup (M1,g); Sup (M2,g);
                     Sdown (M1,g); Sdown (M2,g)]

(* For the non-QCD version we NOW allow gluons and gluinos as external particles.
   (In expectation of Thorsten's marvelous functor.) *)

    let external_flavors () = 
        [ "1st Generation matter", ThoList.flatmap family [1; -1];
          "2nd Generation matter", ThoList.flatmap family [2; -2];
          "3rd Generation matter", ThoList.flatmap family [3; -3];
          "Gauge Bosons", [Ga; Z; Wp; Wm; Gl];
          "Charginos", [Chargino C1; Chargino C2; Chargino C1c; Chargino C2c];
          "Neutralinos", [Neutralino N1; Neutralino N2; Neutralino N3; 
                          Neutralino N4] @
          if Flags.nmssm then [Neutralino N5] else [];
          "Higgs Bosons", [H_Heavy; H_Light; Hp; Hm; A] @
          if Flags.nmssm then [S0; P0] else []; 
          "Gluino", [Gluino]]
          
    let flavors () = ThoList.flatmap snd (external_flavors ())

    let spinor n m =
      if n >= 0 && m >= 0 then
        Spinor
      else if
        n <= 0 && m <=0 then
        ConjSpinor
      else
        invalid_arg "Models.ExtMSSM.spinor: internal error"

    let lorentz = function
      | L g -> spinor g 0 | N g -> spinor g 0
      | U g -> spinor g 0 | D g -> spinor g 0 
      | Chargino c -> spinor (int_of_char c) 0 
      | Ga | Gl -> Vector
      | Wp | Wm | Z -> Massive_Vector
      | H_Heavy | H_Light | Hp | Hm | A -> Scalar
      | Sup _ | Sdown _ | Slepton _ | Sneutrino _  -> Scalar 
      | Neutralino _ -> Majorana 
      | Gluino -> Majorana
      | _ -> invalid_arg "Models.ExtMSSM.lorentz: internal error"

    let color = function
      | U g -> Color.SUN (if g > 0 then 3 else -3)
      | Sup (m,g) -> Color.SUN (if g > 0 then 3 else -3)
      | D g -> Color.SUN (if g > 0 then 3 else -3)
      | Sdown (m,g) -> Color.SUN  (if g > 0 then 3 else -3)
      | Gl | Gluino -> Color.AdjSUN 3
      | _ -> Color.Singlet   


    let prop_spinor n m =
      if n >= 0 && m >=0 then
        Prop_Spinor
      else if 
        n <=0 && m <=0 then
        Prop_ConjSpinor
      else 
        invalid_arg "Models.ExtMSSM.prop_spinor: internal error"

    let propagator = function
      | L g -> prop_spinor g 0 | N g -> prop_spinor g 0
      | U g -> prop_spinor g 0 | D g -> prop_spinor g 0
      | Chargino c -> prop_spinor (int_of_char c) 0 
      | Ga | Gl -> Prop_Feynman
      | Wp | Wm | Z -> Prop_Unitarity
      | H_Heavy | H_Light | Hp | Hm | A | S0 | P0 -> Prop_Scalar
      | Sup _ | Sdown _ | Slepton _ | Sneutrino _ -> Prop_Scalar
      | Gluino -> Prop_Majorana 
(*      | Gluino0 -> Prop_Col_Majorana  *)
      | Neutralino _ -> Prop_Majorana

(* Optionally, ask for the fudge factor treatment for the widths of
   charged particles.  Currently, this only applies to $W^\pm$ and top. *)
                                                                     
    let width f =
      if !use_fudged_width then
        match f with
        | Wp | Wm | U 3 | U (-3) -> Fudged
        | _ -> !default_width
      else
        !default_width 

    let goldstone _ = None

    let conjugate = function
      | L g -> L (-g) | N g -> N (-g)
      | U g -> U (-g) | D g -> D (-g)
      | Sup (m,g) -> Sup (m,-g) 
      | Sdown (m,g) -> Sdown (m,-g) 
      | Slepton (m,g) -> Slepton (m,-g) 
      | Sneutrino g -> Sneutrino (-g)
      | Gl -> Gl | Ga -> Ga | Z -> Z
      | Wp -> Wm | Wm -> Wp
      | H_Heavy -> H_Heavy | H_Light -> H_Light | A -> A
      | Hp -> Hm | Hm -> Hp | S0 -> S0 | P0 -> P0
      | Gluino -> Gluino 
      | Neutralino n -> Neutralino n | Chargino c -> Chargino (conj_char c)
      | _ -> invalid_arg "Models.ExtMSSM.conjugate: internal error"

   let fermion = function
     | L g -> if g > 0 then 1 else -1
     | N g -> if g > 0 then 1 else -1
     | U g -> if g > 0 then 1 else -1
     | D g -> if g > 0 then 1 else -1
     | Gl | Ga | Z | Wp | Wm -> 0 
     | H_Heavy | H_Light | Hp | Hm | A | S0 | P0 -> 0       
     | Neutralino _ -> 2
     | Chargino c -> if (int_of_char c) > 0 then 1 else -1
     | Sup _ -> 0 | Sdown _ -> 0 
     | Slepton _ -> 0 | Sneutrino _ -> 0          
     | Gluino -> 2 

(* Because the O'Caml compiler only allows 248 constructors we must divide the 
   constants into subgroups of constants, e.g. for the Higgs couplings. In the 
   MSSM there are a lot of angles among the parameters, the Weinberg-angle, the 
   angle describing the Higgs vacuum structure, the mixing angle of the real 
   parts of the Higgs dubletts, the mixing angles of the sfermions. Therefore we 
   are going to define the trigonometric functions of those angles not as 
   constants but as functors of the angels. Sums and differences of angles are 
   only used as arguments for the $\alpha$ and $\beta$ angles, so it makes no 
   sense to define special functions for differences and sums of angles. *)

    type angle = 
      | Thw | Al | Be | Th_SF of sff*int | Delta | CKM_12 | CKM_13 | CKM_23

    let string_of_angle = function
      | Thw -> "thw" | Al -> "al" | Be -> "be" | Delta -> "d" 
      | CKM_12 -> "ckm12" | CKM_13 -> "ckm13" | CKM_23 -> "ckm23"
      | Th_SF (f,g) -> "th" ^ string_of_sff f ^ string_of_int g          

(* We introduce a Boolean type vc as a pseudonym for Vertex Conjugator to 
   distinguish between vertices containing complex mixing matrices like the 
   CKM--matrix or the sfermion or neutralino/chargino--mixing matrices, which 
   have to become complex conjugated. The true--option stands for the conjugated 
   vertex, the false--option for the unconjugated vertex. *)

    type vc = bool

    type constant =
      | Unit | Pi | Alpha_QED | Sin2thw
      | Sin of angle | Cos of angle | E | G | Vev | Tanb | Tana 
      | Cos2be | Cos2al | Sin2be | Sin2al | Sin4al | Sin4be | Cos4be
      | Cosapb | Cosamb | Sinapb | Sinamb | Cos2am2b | Sin2am2b       
      | Eidelta
      | Mu | AU of int | AD of int | AL of int 
      | V_CKM of int*int | M_SF of sff*int*sfm*sfm 
      | M_V of char*char  (* left chargino mixing matrix *)
      | M_U of char*char  (* right chargino mixing matrix *)
      | M_N of neu*neu   (* neutralino mixing matrix *)
      | V_0 of neu*neu | A_0 of neu*neu | V_P of char*char | A_P of char*char
      | L_CN of char*neu | R_CN of char*neu | L_NC of neu*char | R_NC of neu*char
(*i      | L_NF of neu*sff*sfm | R_NF of neu*sff*sfm i*)
      | S_NNH1 of neu*neu | P_NNH1 of neu*neu
      | S_NNH2 of neu*neu | P_NNH2 of neu*neu 
      | S_NNA of neu*neu | P_NNA of neu*neu
      | S_NNG of neu*neu | P_NNG of neu*neu
      | L_CNG of char*neu | R_CNG of char*neu
      | L_NCH of neu*char | R_NCH of neu*char
      | Q_lepton | Q_up | Q_down | Q_charg           
      | G_Z | G_CC | G_CCQ of vc*int*int
      | G_NC_neutrino | G_NC_lepton | G_NC_up | G_NC_down 
      | I_Q_W | I_G_ZWW | G_WWWW | G_ZZWW | G_PZWW | G_PPWW 
      | G_strong | G_SS | I_G_S | G_S_Sqrt 
      | Gs
      | M of flavor | W of flavor    
      | G_NZN of neu*neu | G_CZC of char*char 
      | G_YUK of int*int
      | DUM of int
      | G_YUK_1 of int*int | G_YUK_2 of int*int | G_YUK_3 of int*int 
      | G_YUK_4 of int*int | G_NHC of neu*char | G_CHN of char*neu
      | G_YUK_C of vc*int*char*sff*sfm
      | G_YUK_Q of vc*int*int*char*sff*sfm
      | G_YUK_N of vc*int*neu*sff*sfm
      | G_YUK_G of vc*int*sff*sfm
      | G_NGC of neu*char | G_CGN of char*neu 
      | SUM_1 
      | G_NWC of neu*char | G_CWN of char*neu
      | G_CH1C of char*char | G_CH2C of char*char | G_CAC of char*char
      | G_CGC of char*char
      | G_SWS of vc*int*int*sfm*sfm
      | G_SLSNW of vc*int*sfm 
      | G_ZSF of sff*int*sfm*sfm
      | G_CICIH1 of neu*neu | G_CICIH2 of neu*neu | G_CICIA of neu*neu
      | G_CICIG of neu*neu 
      | G_GH of int | G_GHGo of int
      | G_WWSFSF of sff*int*sfm*sfm 
      | G_WPSLSN of vc*int*sfm
      | G_H3 of int | G_H4 of int
      | G_HGo3 of int | G_HGo4 of int | G_GG4 of int
      | G_H1SFSF of sff*int*sfm*sfm | G_H2SFSF of sff*int*sfm*sfm 
      | G_ASFSF of sff*int*sfm*sfm 
      | G_HSNSL of vc*int*sfm  
      | G_GoSFSF of sff*int*sfm*sfm 
      | G_GoSNSL of vc*int*sfm 
      | G_HSUSD of vc*sfm*sfm*int*int | G_GSUSD of vc*sfm*sfm*int*int 
      | G_WPSUSD of vc*sfm*sfm*int*int  
      | G_WZSUSD of vc*sfm*sfm*int*int  
      | G_WZSLSN of vc*int*sfm | G_GlGlSQSQ
      | G_PPSFSF of sff 
      | G_ZZSFSF of sff*int*sfm*sfm | G_ZPSFSF of sff*int*sfm*sfm 
      | G_GlZSFSF of sff*int*sfm*sfm | G_GlPSQSQ 
      | G_GlWSUSD of vc*sfm*sfm*int*int
      | G_GH4 of int | G_GHGo4 of int 
      | G_H1H2SFSF of sff*sfm*sfm*int 
      | G_H1H1SFSF of sff*sfm*sfm*int 
      | G_H2H2SFSF of sff*sfm*sfm*int 
      | G_HHSFSF of sff*sfm*sfm*int 
      | G_AASFSF of sff*sfm*sfm*int 
      | G_HH1SLSN of vc*sfm*int | G_HH2SLSN of vc*sfm*int 
      | G_HASLSN of vc*sfm*int   
      | G_HH1SUSD of vc*sfm*sfm*int*int 
      | G_HH2SUSD of vc*sfm*sfm*int*int 
      | G_HASUSD of vc*sfm*sfm*int*int 
      | G_AG0SFSF of sff*sfm*sfm*int 
      | G_HGSFSF of sff*sfm*sfm*int 
      | G_GGSFSF of sff*sfm*sfm*int 
      | G_G0G0SFSF of sff*sfm*sfm*int
      | G_HGSNSL of vc*sfm*int | G_H1GSNSL of vc*sfm*int 
      | G_H2GSNSL of vc*sfm*int | G_AGSNSL of vc*sfm*int 
      | G_GGSNSL of vc*sfm*int 
      | G_HGSUSD of vc*sfm*sfm*int*int 
      | G_H1GSUSD of vc*sfm*sfm*int*int 
      | G_H2GSUSD of vc*sfm*sfm*int*int 
      | G_AGSUSD of vc*sfm*sfm*int*int 
      | G_GGSUSD of vc*sfm*sfm*int*int 
      | G_SN4 of int*int
      | G_SN2SL2_1 of sfm*sfm*int*int | G_SN2SL2_2 of sfm*sfm*int*int
      | G_SF4 of sff*sff*sfm*sfm*sfm*sfm*int*int
      | G_SF4_3 of sff*sff*sfm*sfm*sfm*sfm*int*int*int
      | G_SF4_4 of sff*sff*sfm*sfm*sfm*sfm*int*int*int*int
      | G_SL4 of sfm*sfm*sfm*sfm*int
      | G_SL4_2 of sfm*sfm*sfm*sfm*int*int
      | G_SN2SQ2 of sff*sfm*sfm*int*int
      | G_SL2SQ2 of sff*sfm*sfm*sfm*sfm*int*int
      | G_SUSDSNSL of vc*sfm*sfm*sfm*int*int*int
      | G_SU4 of sfm*sfm*sfm*sfm*int
      | G_SU4_2 of sfm*sfm*sfm*sfm*int*int
      | G_SD4 of sfm*sfm*sfm*sfm*int
      | G_SD4_2 of sfm*sfm*sfm*sfm*int*int
      | G_SU2SD2 of sfm*sfm*sfm*sfm*int*int*int*int
      | G_HSF31 of higgs*int*sfm*sfm*sff*sff
      | G_HSF32 of higgs*int*int*sfm*sfm*sff*sff
      | G_HSF41 of higgs*int*sfm*sfm*sff*sff
      | G_HSF42 of higgs*int*int*sfm*sfm*sff*sff

    let ferm_of_sff = function
      | SL, g -> (L g) | SN, g -> (N g) 
      | SU, g -> (U g) | SD, g -> (D g)

(* \begin{subequations}
     \begin{align}
        \alpha_{\text{QED}} &= \frac{1}{137.0359895} \\
             \sin^2\theta_w &= 0.23124
     \end{align}
   \end{subequations}

Here we must perhaps allow for complex input parameters. So split them
into their modulus and their phase. At first, we leave them real; the 
generalization to complex parameters is obvious. *)


    let parameters () =
      { input = [];
        derived = [];
        derived_arrays = [] }   

    module F = Models.Fusions (struct
      type f = flavor
      type c = constant
      let compare = compare
      let conjugate = conjugate
    end)


(* For the couplings there are generally two possibilities concerning the
   sign of the covariant derivative. 
   \begin{equation} 
   {\rm CD}^\pm = \partial_\mu \pm \ii g T^a A^a_\mu 
   \end{equation} 
   The particle data group defines the signs consistently to be positive. 
   Since the convention for that signs also influence the phase definitions 
   of the gaugino/higgsino fields via the off-diagonal entries in their
   mass matrices it would be the best to adopt that convention. *)

(*** REVISED: Compatible with CD+. ***)
    let electromagnetic_currents_3 g =
        [ ((L (-g), Ga, L g), FBF (1, Psibar, V, Psi), Q_lepton);
          ((U (-g), Ga, U g), FBF (1, Psibar, V, Psi), Q_up);
          ((D (-g), Ga, D g), FBF (1, Psibar, V, Psi), Q_down)]
        
(*** REVISED: Compatible with CD+. ***)
    let electromagnetic_sfermion_currents g m =
        [ ((Ga, Slepton (m,-g), Slepton (m,g)), Vector_Scalar_Scalar 1, Q_lepton);
          ((Ga, Sup (m,-g), Sup (m,g)), Vector_Scalar_Scalar 1, Q_up);
          ((Ga, Sdown (m,-g), Sdown (m,g)), Vector_Scalar_Scalar 1, Q_down)]       

(*** REVISED: Compatible with CD+. ***)
    let electromagnetic_currents_2 c =
      let cc = conj_char c in
      [ ((Chargino cc, Ga, Chargino c), FBF (1, Psibar, V, Psi), Q_charg) ]

(*** REVISED: Compatible with CD+. ***)
    let neutral_currents g =
      [ ((L (-g), Z, L g), FBF (1, Psibar, VA, Psi), G_NC_lepton);
        ((N (-g), Z, N g), FBF (1, Psibar, VA, Psi), G_NC_neutrino);
        ((U (-g), Z, U g), FBF (1, Psibar, VA, Psi), G_NC_up);
        ((D (-g), Z, D g), FBF (1, Psibar, VA, Psi), G_NC_down)]

(* \begin{equation}
     \mathcal{L}_{\textrm{CC}} =
        \mp \frac{g}{2\sqrt2} \sum_i \bar\psi_i \gamma^\mu
               (1-\gamma_5)(T^+W^+_\mu+T^-W^-_\mu)\psi_i ,
   \end{equation}
   where the sign corresponds to $\text{CD}_\pm$, respectively.  *)

(*** REVISED: Compatible with CD+. ***)
        (* Remark: The definition with the other sign compared to the SM files
           comes from the fact that $g_{cc} = 1/(2\sqrt{2})$ is used 
           overwhelmingly often in the SUSY Feynman rules, so that JR 
           decided to use a different definiton for [g_cc] in SM and MSSM. *)
    let charged_currents g =
      [ ((L (-g), Wm, N g), FBF ((-1), Psibar, VL, Psi), G_CC);
        ((N (-g), Wp, L g), FBF ((-1), Psibar, VL, Psi), G_CC) ]

(* The quark with the inverted generation (the antiparticle) is the outgoing 
   one, the other the incoming. The vertex attached to the outgoing up-quark 
   contains the CKM matrix element {\em not} complex conjugated, while the 
   vertex with the outgoing down-quark has the conjugated CKM matrix 
   element. *)

(*** REVISED: Compatible with CD+. ***)
    let charged_quark_currents g h = 
        [ ((D (-g), Wm, U h), FBF ((-1), Psibar, VL, Psi), G_CCQ (true,g,h));
          ((U (-g), Wp, D h), FBF ((-1), Psibar, VL, Psi), G_CCQ (false,h,g))] 

(*** REVISED: Compatible with CD+. ***)
    let charged_chargino_currents n c =
      let cc = conj_char c in 
      [ ((Chargino cc, Wp, Neutralino n), 
                    FBF (1, Psibar, VLR, Chi), G_CWN (c,n));
        ((Neutralino n, Wm, Chargino c), 
                    FBF (1, Chibar, VLR, Psi), G_NWC (n,c)) ]

(*** REVISED: Compatible with CD+. ***)
    let charged_slepton_currents g m =
      [ ((Wm, Slepton (m,-g), Sneutrino g), Vector_Scalar_Scalar (-1), G_SLSNW 
           (true,g,m));
        ((Wp, Slepton (m,g), Sneutrino (-g)), Vector_Scalar_Scalar 1, G_SLSNW 
           (false,g,m)) ]
 
(*** REVISED: Compatible with CD+. ***)
    let charged_squark_currents' g h m1 m2 =
      [ ((Wm, Sup (m1,g), Sdown (m2,-h)), Vector_Scalar_Scalar (-1), G_SWS 
             (true,g,h,m1,m2));
          ((Wp, Sup (m1,-g), Sdown (m2,h)), Vector_Scalar_Scalar 1, G_SWS 
             (false,g,h,m1,m2)) ]
    let charged_squark_currents g h = List.flatten (Product.list2 
            (charged_squark_currents' g h) [M1;M2] [M1;M2]) 

(*** REVISED: Compatible with CD+. ***)
    let neutral_sfermion_currents' g m1 m2 =
      [ ((Z, Slepton (m1,-g), Slepton (m2,g)), Vector_Scalar_Scalar (-1), G_ZSF 
           (SL,g,m1,m2));
        ((Z, Sup (m1,-g), Sup (m2,g)), Vector_Scalar_Scalar (-1), G_ZSF 
           (SU,g,m1,m2));
        ((Z, Sdown (m1,-g), Sdown (m2,g)), Vector_Scalar_Scalar (-1), G_ZSF 
           (SD,g,m1,m2))]
    let neutral_sfermion_currents g = 
      List.flatten (Product.list2 (neutral_sfermion_currents'
                  g) [M1;M2] [M1;M2]) @
      [ ((Z, Sneutrino (-g), Sneutrino g), Vector_Scalar_Scalar (-1), G_ZSF 
           (SN,g,M1,M1)) ]

(* The reality of the coupling of the Z-boson to two identical neutralinos 
   makes the vector part of the coupling vanish. So we distinguish them not 
   by the name but by the structure of the couplings. *)  

(*** REVISED: Compatible with CD+. ***)
    let neutral_Z_1 (n,m) =  
      [ ((Neutralino n, Z, Neutralino m), FBF (1, Chibar, VA, Chi), 
              (G_NZN (n,m))) ]
(*** REVISED: Compatible with CD+. ***)
    let neutral_Z_2 n =
      [ ((Neutralino n, Z, Neutralino n), FBF (1, Chibar, Coupling.A, Chi), 
              (G_NZN (n,n)) )]

(*** REVISED: Compatible with CD+. ***)
    let charged_Z c1 c2 =
      let cc1 = conj_char c1 in
      ((Chargino cc1, Z, Chargino c2), FBF ((-1), Psibar, VA, Psi), 
               G_CZC (c1,c2)) 

(*** REVISED: Compatible with CD+. 
   Remark: This is pure octet. 
***)        
    let yukawa_v =
      [ (Gluino, Gl, Gluino), FBF (1, Chibar, V, Chi), Gs]

(*** REVISED: Independent of the sign of CD. ***)
    let yukawa_higgs g = 
      [ ((N (-g), Hp, L g), FBF (1, Psibar, Coupling.SR, Psi), G_YUK (6,g));
        ((L (-g), Hm, N g), FBF (1, Psibar, Coupling.SL, Psi), G_YUK (6,g));
        ((L (-g), H_Heavy, L g), FBF (1, Psibar, S, Psi), G_YUK (7,g));
        ((L (-g), H_Light, L g), FBF (1, Psibar, S, Psi), G_YUK (8,g));
        ((L (-g), A, L g), FBF (1, Psibar, P, Psi), G_YUK (9,g));
        ((U (-g), H_Heavy, U g), FBF (1, Psibar, S, Psi), G_YUK (10,g));
        ((U (-g), H_Light, U g), FBF (1, Psibar, S, Psi), G_YUK (11,g));
        ((U (-g), A, U g), FBF (1, Psibar, P, Psi), G_YUK (12,g));
        ((D (-g), H_Heavy, D g), FBF (1, Psibar, S, Psi), G_YUK (13,g));
        ((D (-g), H_Light, D g), FBF (1, Psibar, S, Psi), G_YUK (14,g));
        ((D (-g), A, D g), FBF (1, Psibar, P, Psi), G_YUK (15,g)) ]
        
(*** REVISED: Independent of the sign of CD. ***)
    let yukawa_higgs_quark (g,h) =
      [ ((U (-g), Hp, D h), FBF (1, Psibar, SLR, Psi), G_YUK_1 (g, h)); 
        ((D (-h), Hm, U g), FBF (1, Psibar, SLR, Psi), G_YUK_2 (g, h))  ]

(*** REVISED: Compatible with CD+. *)
    let yukawa_higgs_2' (c1,c2) =
      let cc1 = conj_char c1 in
      [ ((Chargino cc1, H_Heavy, Chargino c2), FBF (1, Psibar, SLR, Psi), 
                G_CH2C (c1,c2));
        ((Chargino cc1, H_Light, Chargino c2), FBF (1, Psibar, SLR, Psi),
                G_CH1C (c1,c2));
        ((Chargino cc1, A, Chargino c2), FBF (1, Psibar, SLR, Psi), 
                G_CAC (c1,c2)) ]
    let yukawa_higgs_2'' c =
      let cc = conj_char c in
      [ ((Chargino cc, H_Heavy, Chargino c), FBF (1, Psibar, S, Psi), 
                G_CH2C (c,c));
        ((Chargino cc, H_Light, Chargino c), FBF (1, Psibar, S, Psi),
                G_CH1C (c,c));
        ((Chargino cc, A, Chargino c), FBF (1, Psibar, P, Psi), 
                G_CAC (c,c)) ]
    let yukawa_higgs_2 = 
      ThoList.flatmap yukawa_higgs_2'  [(C1,C2);(C2,C1)] @ 
      ThoList.flatmap yukawa_higgs_2'' [C1;C2] 

(*** REVISED: Compatible with CD+. ***)
    let higgs_charg_neutr n c =
      let cc = conj_char c in
      [ ((Neutralino n, Hm, Chargino c), FBF (-1, Chibar, SLR, Psi), 
                   G_NHC (n,c));
        ((Chargino cc, Hp, Neutralino n), FBF (-1, Psibar, SLR, Chi), 
                   G_CHN (c,n)) ]

(*** REVISED: Compatible with CD+. ***)        
    let higgs_neutr' (n,m) =
      [ ((Neutralino n, H_Heavy, Neutralino m), FBF (1, Chibar, SP, Chi), 
               G_CICIH2 (n,m));
        ((Neutralino n, H_Light, Neutralino m), FBF (1, Chibar, SP, Chi), 
               G_CICIH1 (n,m));
        ((Neutralino n, A, Neutralino m), FBF (1, Chibar, SP, Chi), 
               G_CICIA (n,m)) ]
    let higgs_neutr'' n =
      [ ((Neutralino n, H_Heavy, Neutralino n), FBF (1, Chibar, S, Chi), 
               G_CICIH2 (n,n));
        ((Neutralino n, H_Light, Neutralino n), FBF (1, Chibar, S, Chi), 
               G_CICIH1 (n,n));
        ((Neutralino n, A, Neutralino n), FBF (1, Chibar, P, Chi), 
               G_CICIA (n,n)) ]
    let higgs_neutr = 
      ThoList.flatmap higgs_neutr'  [(N1,N2);(N1,N3);(N1,N4);
                                     (N2,N3);(N2,N4);(N3,N4)] @ 
      ThoList.flatmap higgs_neutr'' [N1;N2;N3;N4] 
              
(*** REVISED: Compatible with CD+. ***)
     let yukawa_n_1 n g = 
         [ ((Neutralino n, Slepton (M1,-g), L g), FBF (1, Chibar, Coupling.SL,
              Psi), G_YUK_N (true,g,n,SL,M1));
           ((Neutralino n, Slepton (M2,-g), L g), FBF (1, Chibar, SR, Psi), 
              G_YUK_N (true,g,n,SL,M2));
           ((L (-g), Slepton (M1,g), Neutralino n), FBF (1, Psibar, SR, Chi), 
              G_YUK_N (false,g,n,SL,M1));
           ((L (-g), Slepton (M2,g), Neutralino n), FBF (1, Psibar, Coupling.SL,
              Chi), G_YUK_N (false,g,n,SL,M2));
           ((Neutralino n, Sup (M1,-g), U g), FBF (1, Chibar, Coupling.SL, 
              Psi), G_YUK_N (true,g,n,SU,M1));
           ((Neutralino n, Sup (M2,-g), U g), FBF (1, Chibar, SR, Psi), 
              G_YUK_N (true,g,n,SU,M2));
           ((U (-g), Sup (M1,g), Neutralino n), FBF (1, Psibar, SR, Chi), 
              G_YUK_N (false,g,n,SU,M1));
           ((U (-g), Sup (M2,g), Neutralino n), FBF (1, Psibar, Coupling.SL, 
              Chi), G_YUK_N (false,g,n,SU,M2));
           ((Neutralino n, Sdown (M1,-g), D g), FBF (1, Chibar, Coupling.SL, 
              Psi), G_YUK_N (true,g,n,SD,M1));
           ((Neutralino n, Sdown (M2,-g), D g), FBF (1, Chibar, SR, Psi), 
              G_YUK_N (true,g,n,SD,M2));
           ((D (-g), Sdown (M1,g), Neutralino n), FBF (1, Psibar, SR, Chi), 
              G_YUK_N (false,g,n,SD,M1));
           ((D (-g), Sdown (M2,g), Neutralino n), FBF (1, Psibar, Coupling.SL, 
              Chi), G_YUK_N (false,g,n,SD,M2)) ]
     let yukawa_n_2 n m = 
         [ ((Neutralino n, Slepton (m,-3), L 3), FBF (1, Chibar, SLR, Psi), 
              G_YUK_N (true,3,n,SL,m));
           ((L (-3), Slepton (m,3), Neutralino n), FBF (1, Psibar, SLR, Chi), 
              G_YUK_N (false,3,n,SL,m));
           ((Neutralino n, Sup (m,-3), U 3), FBF (1, Chibar, SLR, Psi), 
              G_YUK_N (true,3,n,SU,m));
           ((U (-3), Sup (m,3), Neutralino n), FBF (1, Psibar, SLR, Chi), 
              G_YUK_N (false,3,n,SU,m));
           ((Neutralino n, Sdown (m,-3), D 3), FBF (1, Chibar, SLR, Psi), 
              G_YUK_N (true,3,n,SD,m));
           ((D (-3), Sdown (m,3), Neutralino n), FBF (1, Psibar, SLR, Chi), 
              G_YUK_N (false,3,n,SD,m)) ]
     let yukawa_n_3 n g =
         [ ((Neutralino n, Sneutrino (-g), N g), FBF (1, Chibar, Coupling.SL, 
              Psi), G_YUK_N (true,g,n,SN,M1));
           ((N (-g), Sneutrino g, Neutralino n), FBF (1, Psibar, SR, Chi), 
              G_YUK_N (false,g,n,SN,M1)) ]
(* In principle, we need here the Gluino0 vertices, but they should be automagically 
   generated by Thorsten color script. *)
     let yukawa_n_4 g m =
         [ ((U (-g), Sup (m,g), Gluino), FBF ((-1), Psibar, SR, Chi), G_S_Sqrt);
           ((D (-g), Sdown (m,g), Gluino), FBF ((-1), Psibar, SR, Chi), G_S_Sqrt);
           ((Gluino, Sup (m,-g), U g), FBF ((-1), Chibar, Coupling.SL, Psi), G_S_Sqrt);
           ((Gluino, Sdown (m,-g), D g), FBF ((-1), Chibar, Coupling.SL, Psi), G_S_Sqrt)]
    let yukawa_n_5 m =
          [ ((U (-3), Sup (m,3), Gluino), FBF (1, Psibar, SLR, Chi), 
                G_YUK_G (false,3,SU,m));
           ((D (-3), Sdown (m,3), Gluino), FBF (1, Psibar, SLR, Chi), 
                G_YUK_G (false,3,SD,m));
           ((Gluino, Sup (m,-3), U 3), FBF (1, Chibar, SLR, Psi), 
                G_YUK_G (true,3,SU,m));
           ((Gluino, Sdown (m,-3), D 3), FBF (1, Chibar, SLR, Psi), 
                G_YUK_G (true,3,SD,m))]
    let yukawa_n =
      List.flatten (Product.list2 yukawa_n_1 [N1;N2;N3;N4] [1;2]) @
      List.flatten (Product.list2 yukawa_n_2 [N1;N2;N3;N4] [M1;M2]) @
      List.flatten (Product.list2 yukawa_n_3 [N1;N2;N3;N4] [1;2;3]) @
      List.flatten (Product.list2 yukawa_n_4 [1;2;3] [M1;M2]) @
      ThoList.flatmap yukawa_n_5 [M1;M2]

(*** REVISED: Compatible with CD+. ***)
    let yukawa_c_1 c g =
         let cc = conj_char c in
         [ ((L (-g), Sneutrino g, Chargino cc), BBB (1, Psibar, Coupling.SR, 
              Psibar), G_YUK_C (true,g,c,SN,M1));
           ((Chargino c, Sneutrino (-g), L g), PBP (1, Psi, Coupling.SL, Psi), 
              G_YUK_C (false,g,c,SN,M1)) ]
    let yukawa_c_2 c = 
         let cc = conj_char c in
         [ ((L (-3), Sneutrino 3, Chargino cc), BBB (1, Psibar, SLR, 
              Psibar), G_YUK_C (true,3,c,SN,M1));
           ((Chargino c, Sneutrino (-3), L 3), PBP (1, Psi, SLR, Psi), 
              G_YUK_C (false,3,c,SN,M1)) ]
    let yukawa_c_3 c m g =
         let cc = conj_char c in
         [ ((N (-g), Slepton (m,g), Chargino c), FBF (1, Psibar, Coupling.SR, 
              Psi), G_YUK_C (true,g,c,SL,m));
           ((Chargino cc, Slepton (m,-g), N g), FBF (1, Psibar, Coupling.SL, 
              Psi), G_YUK_C (false,g,c,SL,m)) ]
    let yukawa_c c = 
      ThoList.flatmap (yukawa_c_1 c) [1;2] @ 
      yukawa_c_2 c @
      List.flatten (Product.list2 (yukawa_c_3 c) [M1] [1;2]) @
      List.flatten (Product.list2 (yukawa_c_3 c) [M1;M2] [3]) 

(*** REVISED: Compatible with CD+. ***)
   let yukawa_cq' c (g,h) m = 
       let cc = conj_char c in
         [ ((Chargino c, Sup (m,-g), D h), PBP (1, Psi, SLR, Psi), 
            G_YUK_Q (false,g,h,c,SU,m));
           ((D (-h), Sup (m,g), Chargino cc), BBB (1, Psibar, SLR, Psibar), 
            G_YUK_Q (true,g,h,c,SU,m));
           ((Chargino cc, Sdown (m,-h), U g), FBF (1, Psibar, SLR, Psi), 
            G_YUK_Q (true,g,h,c,SD,m));
           ((U (-g), Sdown (m,h), Chargino c), FBF (1, Psibar, SLR, Psi), 
            G_YUK_Q (false,g,h,c,SD,m)) ]               
   let yukawa_cq'' c (g,h) =
       let cc = conj_char c in
          [ ((Chargino c, Sup (M1,-g), D h), PBP (1, Psi, Coupling.SL, Psi), 
                G_YUK_Q (false,g,h,c,SU,M1));
            ((D (-h), Sup (M1,g), Chargino cc), 
                BBB (1, Psibar, Coupling.SR, Psibar), G_YUK_Q (true,g,h,c,SU,M1));
            ((Chargino cc, Sdown (M1,-h), U g), 
                FBF (1, Psibar, Coupling.SL, Psi), G_YUK_Q (true,g,h,c,SD,M1));
            ((U (-g), Sdown (M1,h), Chargino c), 
                FBF (1, Psibar, Coupling.SR, Psi), G_YUK_Q (false,g,h,c,SD,M1)) ]
   let yukawa_cq c =      
     if Flags.ckm_present then
       List.flatten (Product.list2 (yukawa_cq' c) [(1,3);(2,3);(3,3);
                                                   (3,2);(3,1)] [M1;M2]) @
       ThoList.flatmap (yukawa_cq'' c) [(1,1);(1,2);(2,1);(2,2)] 
     else
       ThoList.flatmap (yukawa_cq' c (3,3)) [M1;M2] @
       ThoList.flatmap (yukawa_cq'' c) [(1,1);(2,2)]


(*** REVISED: Compatible with CD+. 
   Remark: Singlet and octet gluon exchange. The coupling is divided by
   sqrt(2) to account for the correct normalization of the Lie algebra
   generators.
***)         
    let col_currents g =
      [ ((D (-g), Gl, D g), FBF ((-1), Psibar, V, Psi), Gs);
        ((U (-g), Gl, U g), FBF ((-1), Psibar, V, Psi), Gs)]

(*** REVISED: Compatible with CD+. 
   Remark: Singlet and octet gluon exchange. The coupling is divided by
   sqrt(2) to account for the correct normalization of the Lie algebra
   generators.
***)

   let col_sfermion_currents g m = 
      [ ((Gl, Sup (m,-g), Sup (m,g)), Vector_Scalar_Scalar (-1), Gs);
        ((Gl, Sdown (m,-g), Sdown (m,g)), Vector_Scalar_Scalar (-1), Gs)]

(*** REVISED: Compatible with CD+. ***)
   let triple_gauge =
      [ ((Ga, Wm, Wp), Gauge_Gauge_Gauge 1, I_Q_W);  
        ((Z, Wm, Wp), Gauge_Gauge_Gauge 1, I_G_ZWW);
        ((Gl, Gl, Gl), Gauge_Gauge_Gauge 1, I_G_S)]

(*** REVISED: Independent of the sign of CD. ***) 
   let gauge4 = Vector4 [(2, C_13_42); (-1, C_12_34); (-1, C_14_23)]
   let minus_gauge4 = Vector4 [(-2, C_13_42); (1, C_12_34); (1, C_14_23)] 
   let quartic_gauge =
      [ (Wm, Wp, Wm, Wp), gauge4, G_WWWW;
        (Wm, Z, Wp, Z), minus_gauge4, G_ZZWW;
        (Wm, Z, Wp, Ga), minus_gauge4, G_PZWW;
        (Wm, Ga, Wp, Ga), minus_gauge4, G_PPWW;
        (Gl, Gl, Gl, Gl), gauge4, G_SS]

(* The [Scalar_Vector_Vector] couplings do not depend on the choice of the
   sign of the covariant derivative since they are quadratic in the
   gauge couplings. *)

(*** REVISED: Compatible with CD+. ***)
(*** Revision: 2005-03-10: first two vertices corrected. ***)
    let gauge_higgs =
      [ ((Wm, Hp, A), Vector_Scalar_Scalar 1, G_GH 1);
        ((Wp, Hm, A), Vector_Scalar_Scalar 1, G_GH 1);
        ((Z, H_Heavy, A), Vector_Scalar_Scalar 1, G_GH 3);
        ((Z, H_Light, A), Vector_Scalar_Scalar 1, G_GH 2);
        ((H_Heavy, Wp, Wm), Scalar_Vector_Vector 1, G_GH 5);
        ((H_Light, Wp, Wm), Scalar_Vector_Vector 1, G_GH 4);
        ((Wm, Hp, H_Heavy), Vector_Scalar_Scalar 1, G_GH 7);
        ((Wp, Hm, H_Heavy), Vector_Scalar_Scalar (-1), G_GH 7);
        ((Wm, Hp, H_Light), Vector_Scalar_Scalar 1, G_GH 6);
        ((Wp, Hm, H_Light), Vector_Scalar_Scalar (-1), G_GH 6);        
        ((H_Heavy, Z, Z), Scalar_Vector_Vector 1, G_GH 9);
        ((H_Light, Z, Z), Scalar_Vector_Vector 1, G_GH 8);
        ((Z, Hp, Hm), Vector_Scalar_Scalar 1, G_GH 10);
        ((Ga, Hp, Hm), Vector_Scalar_Scalar 1, G_GH 11) ]

    let gauge_higgs4 = 
      [ ((A, A, Z, Z), Scalar2_Vector2 1, G_GH4 1);
        ((H_Heavy, H_Heavy, Z, Z), Scalar2_Vector2 1, G_GH4 3);
        ((H_Light, H_Light, Z, Z), Scalar2_Vector2 1, G_GH4 2);
        ((Hp, Hm, Z, Z), Scalar2_Vector2 1, G_GH4 4);
        ((Hp, Hm, Ga, Ga), Scalar2_Vector2 1, G_GH4 5);
        ((Hp, Hm, Ga, Z), Scalar2_Vector2 1, G_GH4 6);
        ((Hp, H_Heavy, Wm, Z), Scalar2_Vector2 1, G_GH4 8);
        ((Hm, H_Heavy, Wp, Z), Scalar2_Vector2 1, G_GH4 8);
        ((Hp, H_Light, Wm, Z), Scalar2_Vector2 1, G_GH4 7);
        ((Hm, H_Light, Wp, Z), Scalar2_Vector2 1, G_GH4 7);
        ((Hp, H_Heavy, Wm, Ga), Scalar2_Vector2 1, G_GH4 10);
        ((Hm, H_Heavy, Wp, Ga), Scalar2_Vector2 1, G_GH4 10);
        ((Hp, H_Light, Wm, Ga), Scalar2_Vector2 1, G_GH4 9);
        ((Hm, H_Light, Wp, Ga), Scalar2_Vector2 1, G_GH4 9);
        ((A, A, Wp, Wm), Scalar2_Vector2 1, G_GH4 11); 
        ((H_Heavy, H_Heavy, Wp, Wm), Scalar2_Vector2 1, G_GH4 13);
        ((H_Light, H_Light, Wp, Wm), Scalar2_Vector2 1, G_GH4 12);
        ((Hp, Hm, Wp, Wm), Scalar2_Vector2 1, G_GH4 14);
        ((Hp, A, Wm, Z), Scalar2_Vector2 1, G_GH4 15);
        ((Hm, A, Wp, Z), Scalar2_Vector2 (-1), G_GH4 15);
        ((Hp, A, Wm, Ga), Scalar2_Vector2 1, G_GH4 16);
        ((Hm, A, Wp, Ga), Scalar2_Vector2 (-1), G_GH4 16) ]

    let gauge_sfermion4' g m1 m2 =
       [ ((Wp, Wm, Slepton (m1,g), Slepton (m2,-g)), Scalar2_Vector2 1, 
            G_WWSFSF (SL,g,m1,m2));
        ((Z, Ga, Slepton (m1,g), Slepton (m2,-g)), Scalar2_Vector2 1, 
           G_ZPSFSF (SL,g,m1,m2));
        ((Z, Z, Slepton (m1,g), Slepton (m2,-g)), Scalar2_Vector2 1, G_ZZSFSF 
           (SL,g,m1,m2)); 
        ((Wp, Wm, Sup (m1,g), Sup (m2,-g)), Scalar2_Vector2 1, G_WWSFSF 
           (SU,g,m1,m2));
        ((Wp, Wm, Sdown (m1,g), Sdown (m2,-g)), Scalar2_Vector2 1, G_WWSFSF 
           (SD,g,m1,m2));
        ((Z, Z, Sup (m1,g), Sup (m2,-g)), Scalar2_Vector2 1, G_ZZSFSF 
           (SU,g,m1,m2));
        ((Z, Z, Sdown (m1,g), Sdown (m2,-g)), Scalar2_Vector2 1, G_ZZSFSF 
           (SD,g,m1,m2));
        ((Z, Ga, Sup (m1,g), Sup (m2,-g)), Scalar2_Vector2 1, G_ZPSFSF 
           (SU,g,m1,m2));
        ((Z, Ga, Sdown (m1,g), Sdown (m2,-g)), Scalar2_Vector2 1, G_ZPSFSF 
           (SD,g,m1,m2)) ]
    let gauge_sfermion4'' g m =
      [ ((Wp, Ga, Slepton (m,g), Sneutrino (-g)), Scalar2_Vector2 1, G_WPSLSN 
           (false,g,m));
        ((Wm, Ga, Slepton (m,-g), Sneutrino g), Scalar2_Vector2 1, 
           G_WPSLSN (true,g,m));
        ((Wp, Z, Slepton (m,g), Sneutrino (-g)), Scalar2_Vector2 1, G_WZSLSN 
           (false,g,m));
        ((Wm, Z, Slepton (m,-g), Sneutrino g), Scalar2_Vector2 1,
           G_WZSLSN (true,g,m));
        ((Ga, Ga, Slepton (m,g), Slepton (m,-g)), Scalar2_Vector2 1, G_PPSFSF SL); 
        ((Ga, Ga, Sup (m,g), Sup (m,-g)), Scalar2_Vector2 1, G_PPSFSF SU);
        ((Ga, Ga, Sdown (m,g), Sdown (m,-g)), Scalar2_Vector2 1, G_PPSFSF SD)]
    let gauge_sfermion4 g =
      List.flatten (Product.list2 (gauge_sfermion4' g) [M1;M2] [M1;M2]) @
      ThoList.flatmap (gauge_sfermion4'' g) [M1;M2] @
      [ ((Wp, Wm, Sneutrino g, Sneutrino (-g)), Scalar2_Vector2 1, G_WWSFSF 
           (SN,g,M1,M1));
        ((Z, Z, Sneutrino g, Sneutrino (-g)), Scalar2_Vector2 1, G_ZZSFSF 
           (SN,g,M1,M1)) ]

    let gauge_squark4' g h m1 m2 =
      [ ((Wp, Ga, Sup (m1,-g), Sdown (m2,h)), Scalar2_Vector2 1, G_WPSUSD 
           (false,m1,m2,g,h));
        ((Wm, Ga, Sup (m1,g), Sdown (m2,-h)), Scalar2_Vector2 1, G_WPSUSD 
           (true,m1,m2,g,h));
        ((Wp, Z, Sup (m1,-g), Sdown (m2,h)), Scalar2_Vector2 1, G_WZSUSD 
           (false,m1,m2,g,h));
        ((Wm, Z, Sup (m1,g), Sdown (m2,-h)), Scalar2_Vector2 1, G_WZSUSD 
           (true,m1,m2,g,h)) ]
      let gauge_squark4 g h = List.flatten (Product.list2 (gauge_squark4' g h) 
                                              [M1;M2] [M1;M2])

    let gluon_w_squark' g h m1 m2 =
      [ ((Gl, Wp, Sup (m1,-g), Sdown (m2,h)), 
            Scalar2_Vector2 1, G_GlWSUSD (false,m1,m2,g,h));
        ((Gl, Wm, Sup (m1,g), Sdown (m2,-h)), 
            Scalar2_Vector2 1, G_GlWSUSD (true,m1,m2,g,h)) ]
    let gluon_w_squark g h = 
      List.flatten (Product.list2 (gluon_w_squark' g h) [M1;M2] [M1;M2])

    let gluon_gauge_squark' g m1 m2 =
      [ ((Gl, Z, Sup (m1,g), Sup (m2,-g)), 
            Scalar2_Vector2 2, G_GlZSFSF (SU,g,m1,m2));
        ((Gl, Z, Sdown (m1,g), Sdown (m2,-g)), 
            Scalar2_Vector2 2, G_GlZSFSF (SD,g,m1,m2)) ]
    let gluon_gauge_squark'' g m =
      [ ((Gl, Ga, Sup (m,g), Sup (m,-g)), Scalar2_Vector2 2, G_GlPSQSQ);
        ((Gl, Ga, Sdown (m,g), Sdown (m,-g)), Scalar2_Vector2 (-1), G_GlPSQSQ) ]
    let gluon_gauge_squark g =
      List.flatten (Product.list2 (gluon_gauge_squark' g) [M1;M2] [M1;M2]) @
      ThoList.flatmap (gluon_gauge_squark'' g) [M1;M2]

    let gluon2_squark2' g m = 
      [ ((Gl, Gl, Sup (m,g), Sup (m,-g)), Scalar2_Vector2 2, G_GlGlSQSQ);
        ((Gl, Gl, Sdown (m,g), Sdown (m,-g)), Scalar2_Vector2 2, G_GlGlSQSQ) ] 
    let gluon2_squark2 g = 
      ThoList.flatmap (gluon2_squark2' g) [M1;M2] 


(*** REVISED: Independent of the sign of CD. ***)
    let higgs =
      [ ((Hp, Hm, H_Heavy), Scalar_Scalar_Scalar 1, G_H3 1);
        ((Hp, Hm, H_Light), Scalar_Scalar_Scalar 1, G_H3 2);
        ((H_Heavy, H_Heavy, H_Light), Scalar_Scalar_Scalar 1, G_H3 3);
        ((H_Heavy, H_Heavy, H_Heavy), Scalar_Scalar_Scalar 1, G_H3 4);
        ((H_Light, H_Light, H_Light), Scalar_Scalar_Scalar 1, G_H3 5);
        ((H_Heavy, H_Light, H_Light), Scalar_Scalar_Scalar 1, G_H3 6);
        ((H_Heavy, A, A), Scalar_Scalar_Scalar 1, G_H3 7);
        ((H_Light, A, A), Scalar_Scalar_Scalar 1, G_H3 8) ]

(* Here follow purely scalar quartic vertices which are only available for the
   no-Whizard colored version. *)

(*** REVISED: Independent of the sign of CD. ***)
    let higgs4 =
      [ ((Hp, Hm, Hp, Hm), Scalar4 1, G_H4 1);
        ((Hp, Hm, H_Heavy, H_Heavy), Scalar4 1, G_H4 2);
        ((Hp, Hm, H_Light, H_Light), Scalar4 1, G_H4 3);
        ((Hp, Hm, H_Heavy, H_Light), Scalar4 1, G_H4 4);
        ((Hp, Hm, A, A), Scalar4 1, G_H4 5);
        ((H_Heavy, H_Heavy, H_Heavy, H_Heavy), Scalar4 1, G_H4 6);
        ((H_Light, H_Light, H_Light, H_Light), Scalar4 1, G_H4 6);
        ((H_Heavy, H_Heavy, H_Light, H_Light), Scalar4 1, G_H4 7);
        ((H_Heavy, H_Light, H_Light, H_Light), Scalar4 1, G_H4 8);
        ((H_Heavy, H_Heavy, H_Heavy, H_Light), Scalar4 (-1), G_H4 8);
        ((H_Heavy, H_Heavy, A, A), Scalar4 1, G_H4 9);
        ((H_Light, H_Light, A, A), Scalar4 (-1), G_H4 9);
        ((H_Heavy, H_Light, A, A), Scalar4 1, G_H4 10);
        ((A, A, A, A), Scalar4 1, G_H4 11) ]
    
(* The vertices of the type Higgs - Sfermion - Sfermion are independent of 
   the choice of the CD sign since they are quadratic in the gauge 
   coupling. *) 

(*** REVISED: Independent of the sign of CD. ***)
    let higgs_sneutrino' g =
      [ ((H_Heavy, Sneutrino g, Sneutrino (-g)), Scalar_Scalar_Scalar 1, 
                       G_H2SFSF (SN,g,M1,M1));
        ((H_Light, Sneutrino g, Sneutrino (-g)), Scalar_Scalar_Scalar 1, 
                       G_H1SFSF (SN,g,M1,M1));
        ((Hp, Sneutrino (-g), Slepton (M1,g)), Scalar_Scalar_Scalar 1, 
              G_HSNSL (false,g,M1)); 
        ((Hm, Sneutrino g, Slepton (M1,-g)), Scalar_Scalar_Scalar 1, 
              G_HSNSL (true,g,M1)) ]
      let higgs_sneutrino'' = 
      [ ((Hp, Sneutrino (-3), Slepton (M2,3)), Scalar_Scalar_Scalar 1, 
              G_HSNSL (false,3,M2)); 
        ((Hm, Sneutrino 3, Slepton (M2,-3)), Scalar_Scalar_Scalar 1, 
              G_HSNSL (false,3,M2)) ]
      let higgs_sneutrino = 
        ThoList.flatmap higgs_sneutrino' [1;2;3] @ higgs_sneutrino''  
        

(* Under the assumption that there is no mixing between the left- and
   right-handed sfermions for the first two generations there is only a 
   coupling of the form Higgs - sfermion1 - sfermion2 for the third 
   generation. All the others are suppressed by $m_f/M_W$. *)

(*** REVISED: Independent of the sign of CD. ***)
      let higgs_sfermion' g m1 m2 =
        [ ((H_Heavy, Slepton (m1,g), Slepton (m2,-g)), Scalar_Scalar_Scalar 1,
              G_H2SFSF (SL,g,m1,m2));
          ((H_Light, Slepton (m1,g), Slepton (m2,-g)), Scalar_Scalar_Scalar 1,
              G_H1SFSF (SL,g,m1,m2));
          ((H_Heavy, Sup (m1,g), Sup (m2,-g)), Scalar_Scalar_Scalar 1, 
              G_H2SFSF (SU,g,m1,m2));
          ((H_Heavy, Sdown (m1,g), Sdown (m2,-g)), Scalar_Scalar_Scalar 1, 
              G_H2SFSF (SD,g,m1,m2));
          ((H_Light, Sup (m1,g), Sup (m2,-g)), Scalar_Scalar_Scalar 1, 
              G_H1SFSF (SU,g,m1,m2));
          ((H_Light, Sdown (m1,g), Sdown (m2,-g)), Scalar_Scalar_Scalar 1, 
              G_H1SFSF (SD,g,m1,m2)) ]
      let higgs_sfermion'' m1 m2 = 
        [ ((A, Slepton (m1,3), Slepton (m2,-3)), Scalar_Scalar_Scalar 1,
              G_ASFSF (SL,3,m1,m2));
          ((A, Sup (m1,3), Sup (m2,-3)), Scalar_Scalar_Scalar 1, 
              G_ASFSF (SU,3,m1,m2));
          ((A, Sdown (m1,3), Sdown (m2,-3)), Scalar_Scalar_Scalar 1, 
              G_ASFSF (SD,3,m1,m2)) ]
    let higgs_sfermion = List.flatten (Product.list2 (higgs_sfermion' 3) 
                                           [M1;M2] [M1;M2]) @ 
        (higgs_sfermion' 1 M1 M1) @ (higgs_sfermion' 1 M2 M2) @
        (higgs_sfermion' 2 M1 M1) @ (higgs_sfermion' 2 M2 M2) @ 
        List.flatten (Product.list2 higgs_sfermion'' [M1;M2] [M1;M2]) 
   
(*i    let higgs_sfermion g = List.flatten (Product.list2 (higgs_sfermion' g) 
                                           [M1;M2] [M1;M2])  i*)

(*** REVISED: Independent of the sign of CD. ***)
    let higgs_squark' g h m1 m2 =
      [ ((Hp, Sup (m1,-g), Sdown (m2,h)), Scalar_Scalar_Scalar 1, 
              G_HSUSD (false,m1,m2,g,h)); 
        ((Hm, Sup (m1,g), Sdown (m2,-h)), Scalar_Scalar_Scalar 1, 
              G_HSUSD (true,m1,m2,g,h)) ]
    let higgs_squark_a g h = higgs_squark' g h M1 M1 
    let higgs_squark_b (g,h) = List.flatten (Product.list2 (higgs_squark' g h)
                                             [M1;M2] [M1;M2]) 
    let higgs_squark =          
         List.flatten (Product.list2 higgs_squark_a [1;2] [1;2]) @ 
         ThoList.flatmap higgs_squark_b [(1,3);(2,3);(3,3);(3,1);(3,2)] 

    let vertices3 = 
        (ThoList.flatmap electromagnetic_currents_3 [1;2;3] @
         ThoList.flatmap electromagnetic_currents_2 [C1;C2] @
         List.flatten (Product.list2 electromagnetic_sfermion_currents [1;2;3] 
                         [M1;M2]) @ 
         ThoList.flatmap neutral_currents [1;2;3] @
         ThoList.flatmap neutral_sfermion_currents [1;2;3] @  
         ThoList.flatmap charged_currents [1;2;3] @
         List.flatten (Product.list2 charged_slepton_currents [1;2;3] 
                         [M1;M2]) @ 
         (if Flags.ckm_present then 
           List.flatten (Product.list2 charged_quark_currents [1;2;3] 
                           [1;2;3]) @
           List.flatten (Product.list2 charged_squark_currents [1;2;3] 
                           [1;2;3]) @ 
           ThoList.flatmap yukawa_higgs_quark [(1,3);(2,3);(3,3);(3,1);(3,2)]
         else
           charged_quark_currents 1 1 @
           charged_quark_currents 2 2 @
           charged_quark_currents 3 3 @
           charged_squark_currents 1 1 @
           charged_squark_currents 2 2 @
           charged_squark_currents 3 3 @ 
           ThoList.flatmap yukawa_higgs_quark [(3,3)]) @ 
(*i         ThoList.flatmap yukawa_higgs [1;2;3] @  i*)
         yukawa_higgs 3 @ yukawa_n @ 
         ThoList.flatmap yukawa_c [C1;C2] @ 
         ThoList.flatmap yukawa_cq [C1;C2] @ 
         List.flatten (Product.list2 charged_chargino_currents [N1;N2;N3;N4] 
                         [C1;C2]) @ triple_gauge @ 
         ThoList.flatmap neutral_Z_1 [(N1,N2);(N1,N3);(N1,N4);(N2,N3);(N2,N4);
                                    (N3,N4)] @
         ThoList.flatmap neutral_Z_2 [N1;N2;N3;N4] @
         Product.list2 charged_Z [C1;C2] [C1;C2] @ 
         gauge_higgs @ higgs @ yukawa_higgs_2 @ 
(*i         List.flatten (Product.list2 yukawa_higgs_quark [1;2;3] [1;2;3]) @  i*)
         List.flatten (Product.list2 higgs_charg_neutr [N1;N2;N3;N4] [C1;C2]) @ 
         higgs_neutr @ higgs_sneutrino @ higgs_sfermion @ 
(*i         ThoList.flatmap higgs_sfermion [1;2;3] @  i*)
         higgs_squark @ yukawa_v @
         ThoList.flatmap col_currents [1;2;3] @
         List.flatten (Product.list2 col_sfermion_currents [1;2;3] [M1;M2])) 

(*    let vertices4 = []  *) 

    let vertices4 =
       (quartic_gauge @ higgs4 @ gauge_higgs4 @ 
        ThoList.flatmap gauge_sfermion4 [1;2;3] @
        List.flatten (Product.list2 gauge_squark4 [1;2;3] [1;2;3]) @
        ThoList.flatmap gluon2_squark2  [1;2;3] @
        List.flatten (Product.list2 gluon_w_squark [1;2;3] [1;2;3]) @
        ThoList.flatmap gluon_gauge_squark [1;2;3])
         
    let vertices () = (vertices3, vertices4, [])

    let table = F.of_vertices (vertices ())
    let fuse2 = F.fuse2 table
    let fuse3 = F.fuse3 table
    let fuse = F.fuse table                                    
    let max_degree () = 4

    let flavor_of_string s = 
      match s with
          | "e-" -> L 1 | "e+" -> L (-1)
          | "mu-" -> L 2 | "mu+" -> L (-2)
          | "tau-" -> L 3 | "tau+" -> L (-3)
          | "nue" -> N 1 | "nuebar" -> N (-1)
          | "numu" -> N 2 | "numubar" -> N (-2)
          | "nutau" -> N 3 | "nutaubar" -> N (-3)
          | "se1-" -> Slepton (M1,1) | "se1+" -> Slepton (M1,-1)
          | "smu1-" -> Slepton (M1,2) | "smu1+" -> Slepton (M1,-2)
          | "stau1-" -> Slepton (M1,3) | "stau1+" -> Slepton (M1,-3)
          | "se2-" -> Slepton (M2,1) | "se2+" -> Slepton (M2,-1)
          | "smu2-" -> Slepton (M2,2) | "smu2+" -> Slepton (M2,-2)
          | "stau2-" -> Slepton (M2,3) | "stau2+" -> Slepton (M2,-3)
          | "snue" -> Sneutrino 1 | "snue*" -> Sneutrino (-1)
          | "snumu" -> Sneutrino 2 | "snumu*" -> Sneutrino (-2)
          | "snutau" -> Sneutrino 3 | "snutau*" -> Sneutrino (-3)
          | "u" -> U 1 | "ubar" -> U (-1)
          | "c" -> U 2 | "cbar" -> U (-2)
          | "t" -> U 3 | "tbar" -> U (-3)
          | "d" -> D 1 | "dbar" -> D (-1)
          | "s" -> D 2 | "sbar" -> D (-2)
          | "b" -> D 3 | "bbar" -> D (-3)
          | "A" -> Ga | "Z" | "Z0" -> Z
          | "W+" -> Wp | "W-" -> Wm
          | "H" -> H_Heavy | "h" -> H_Light | "A0" -> A 
          | "H+" -> Hp | "H-" -> Hm
          | "su1" -> Sup (M1,1) | "su1c" -> Sup (M1,-1)
          | "sc1" -> Sup (M1,2) | "sc1c" -> Sup (M1,-2)
          | "st1" -> Sup (M1,3) | "st1c" -> Sup (M1,-3)
          | "su2" -> Sup (M2,1) | "su2c" -> Sup (M2,-1)
          | "sc2" -> Sup (M2,2) | "sc2c" -> Sup (M2,-2)
          | "st2" -> Sup (M2,3) | "st2c" -> Sup (M2,-3)
          | "sd1" -> Sdown (M1,1) | "sd1c" -> Sdown (M1,-1)
          | "ss1" -> Sdown (M1,2) | "ss1c" -> Sdown (M1,-2)
          | "sb1" -> Sdown (M1,3) | "sb1c" -> Sdown (M1,-3)
          | "sd2" -> Sdown (M2,1) | "sd2c" -> Sdown (M2,-1)
          | "ss2" -> Sdown (M2,2) | "ss2c" -> Sdown (M2,-2)
          | "sb2" -> Sdown (M2,3) | "sb2c" -> Sdown (M2,-3)
          | "neu1" -> Neutralino N1 | "neu2" -> Neutralino N2
          | "neu3" -> Neutralino N3 | "neu4" -> Neutralino N4      
          | "neu5" -> if Flags.nmssm then Neutralino N5 else
            invalid_arg "Models.ExtMSSM.flavor_of_string"
          | "ch1+" -> Chargino C1 | "ch2+" -> Chargino C2
          | "ch1-" -> Chargino C1c | "ch2-" -> Chargino C2c
          | _ -> invalid_arg "Models.ExtMSSM.flavor_of_string"
                
    let flavor_to_string = function
      | L 1 -> "e-" | L (-1) -> "e+"
      | L 2 -> "mu-" | L (-2) -> "mu+"
      | L 3 -> "tau-" | L (-3) -> "tau+"
      | N 1 -> "nue" | N (-1) -> "nuebar"
      | N 2 -> "numu" | N (-2) -> "numubar"
      | N 3 -> "nutau" | N (-3) -> "nutaubar"
      | U 1 -> "u" | U (-1) -> "ubar"
      | U 2 -> "c" | U (-2) -> "cbar"
      | U 3 -> "t" | U (-3) -> "tbar"
      | U _ -> invalid_arg
            "Models.ExtMSSM.flavor_to_string: invalid up type quark"
      | D 1 -> "d" | D (-1) -> "dbar"
      | D 2 -> "s" | D (-2) -> "sbar"
      | D 3 -> "b" | D (-3) -> "bbar"
      | D _ -> invalid_arg
            "Models.ExtMSSM.flavor_to_string: invalid down type quark"
      | Gl -> "gl" | Gluino -> "sgl"
      | Ga -> "A" | Z -> "Z" 
      | Wp -> "W+" | Wm -> "W-"
      | H_Heavy -> "H" | H_Light -> "h" | A -> "A0" 
      | Hp -> "H+" | Hm -> "H-"
      | S0 -> "S0" | P0 -> "P0"
      | Slepton (M1,1) -> "se1-" | Slepton (M1,-1) -> "se1+"
      | Slepton (M1,2) -> "smu1-" | Slepton (M1,-2) -> "smu1+"
      | Slepton (M1,3) -> "stau1-" | Slepton (M1,-3) -> "stau1+"
      | Slepton (M2,1) -> "se2-" | Slepton (M2,-1) -> "se2+"
      | Slepton (M2,2) -> "smu2-" | Slepton (M2,-2) -> "smu2+"
      | Slepton (M2,3) -> "stau2-" | Slepton (M2,-3) -> "stau2+"
      | Sneutrino 1 -> "snue" | Sneutrino (-1) -> "snue*"
      | Sneutrino 2 -> "snumu" | Sneutrino (-2) -> "snumu*"
      | Sneutrino 3 -> "snutau" | Sneutrino (-3) -> "snutau*"
      | Sup (M1,1) -> "su1+" | Sup (M1,-1) -> "su1-"
      | Sup (M1,2) -> "sc1+" | Sup (M1,-2) -> "sc1-"
      | Sup (M1,3) -> "st1+" | Sup (M1,-3) -> "st1-"
      | Sup (M2,1) -> "su2+" | Sup (M2,-1) -> "su2-"
      | Sup (M2,2) -> "sc2+" | Sup (M2,-2) -> "sc2-"
      | Sup (M2,3) -> "st2+" | Sup (M2,-3) -> "st2-"
      | Sdown (M1,1) -> "sd1-" | Sdown (M1,-1) -> "sd1+"
      | Sdown (M1,2) -> "ss1-" | Sdown (M1,-2) -> "ss1+"
      | Sdown (M1,3) -> "sb1-" | Sdown (M1,-3) -> "sb1+"
      | Sdown (M2,1) -> "sd2-" | Sdown (M2,-1) -> "sd2+"
      | Sdown (M2,2) -> "ss2-" | Sdown (M2,-2) -> "ss2+"
      | Sdown (M2,3) -> "sb2-" | Sdown (M2,-3) -> "sb2+"
      | Neutralino N1 -> "neu1"
      | Neutralino N2 -> "neu2"
      | Neutralino N3 -> "neu3"
      | Neutralino N4 -> "neu4"
      | Neutralino N5 -> if Flags.nmssm then "neu5" else
        invalid_arg "Models.ExtMSSM.flavor_to_string"
      | Chargino C1 -> "ch1+" | Chargino C1c -> "ch1-"
      | Chargino C2 -> "ch2+" | Chargino C2c -> "ch2-"
      | _ -> invalid_arg "Models.ExtMSSM.flavor_to_string"

    let flavor_symbol = function
      | L g when g > 0 -> "l" ^ string_of_int g
      | L g -> "l" ^ string_of_int (abs g) ^ "b"  
      | N g when g > 0 -> "n" ^ string_of_int g
      | N g -> "n" ^ string_of_int (abs g) ^ "b"      
      | U g when g > 0 -> "u" ^ string_of_int g 
      | U g -> "u" ^ string_of_int (abs g) ^ "b"  
      | D g when g > 0 ->  "d" ^ string_of_int g 
      | D g -> "d" ^ string_of_int (abs g) ^ "b"    
      | Gl -> "gl" 
      | Ga -> "a" | Z -> "z"
      | Wp -> "wp" | Wm -> "wm"
      | Slepton (M1,g) when g > 0 -> "sl1" ^ string_of_int g 
      | Slepton (M1,g) -> "sl1c" ^ string_of_int (abs g)
      | Slepton (M2,g) when g > 0 -> "sl2" ^ string_of_int g
      | Slepton (M2,g) -> "sl2c" ^ string_of_int (abs g)
      | Sneutrino g when g > 0 -> "sn" ^ string_of_int g
      | Sneutrino g -> "snc" ^ string_of_int (abs g)
      | Sup (M1,g) when g > 0 -> "su1" ^ string_of_int g
      | Sup (M1,g) -> "su1c" ^ string_of_int (abs g)
      | Sup (M2,g) when g > 0 -> "su2" ^ string_of_int g
      | Sup (M2,g) -> "su2c" ^ string_of_int (abs g)
      | Sdown (M1,g) when g > 0 ->  "sd1" ^ string_of_int g
      | Sdown (M1,g) -> "sd1c" ^ string_of_int (abs g)
      | Sdown (M2,g) when g > 0 ->  "sd2" ^ string_of_int g
      | Sdown (M2,g) -> "sd2c" ^ string_of_int (abs g)
      | Neutralino n -> "neu" ^ (string_of_neu n)
      | Chargino c when (int_of_char c) > 0 -> "cp" ^ string_of_char c
      | Chargino c -> "cm" ^ string_of_int (abs (int_of_char c))
      | Gluino -> "sgl" 
      | H_Heavy -> "h0h" | H_Light -> "h0l" | A -> "a0"
      | Hp -> "hp" | Hm -> "hm" | S0 -> "s0" | P0 -> "p0"

     let pdg = function
      | L g when g > 0 -> 9 + 2*g
      | L g -> - 9 + 2*g
      | N g when g > 0 -> 10 + 2*g
      | N g -> - 10 + 2*g
      | U g  when g > 0 -> 2*g
      | U g  -> 2*g
      | D g  when g > 0 -> - 1 + 2*g
      | D g  -> 1 + 2*g
      | Gl -> 21 
      | Ga -> 22 | Z -> 23
      | Wp -> 24 | Wm -> (-24)
      | H_Light -> 25 | H_Heavy -> 35 | A -> 36
      | Hp -> 37 | Hm -> (-37)
      | Slepton (M1,g) when g > 0 -> 1000009 + 2*g
      | Slepton (M1,g) -> - 1000009 + 2*g
      | Slepton (M2,g) when g > 0 -> 2000009 + 2*g
      | Slepton (M2,g) -> - 2000009 + 2*g            
      | Sneutrino g when g > 0 -> 1000010 + 2*g
      | Sneutrino g -> - 1000010 + 2*g            
      | Sup (M1,g) when g > 0 -> 1000000 + 2*g
      | Sup (M1,g) -> - 1000000 + 2*g
      | Sup (M2,g) when g > 0 -> 2000000 + 2*g
      | Sup (M2,g) -> - 2000000 + 2*g
      | Sdown (M1,g) when g > 0 -> 999999 + 2*g
      | Sdown (M1,g) -> - 999999 + 2*g
      | Sdown (M2,g) when g > 0 -> 1999999 + 2*g
      | Sdown (M2,g) -> - 1999999 + 2*g
      | Gluino -> 1000021
      | Chargino C1 -> 1000024 | Chargino C1c -> (-1000024)
      | Chargino C2 -> 1000037 | Chargino C2c -> (-1000037)
      | Neutralino N1 -> 1000022 | Neutralino N2 -> 1000023
      | Neutralino N3 -> 1000025 | Neutralino N4 -> 1000035
(* Bitte hier noch die richtigen Werte einsetzen. *)
      | Neutralino N5 -> 1000099
      | P0 -> 1111111 | S0 -> 1111112


(* We must take care of the pdg numbers for the two different kinds of 
   sfermions in the MSSM. The particle data group in its Monte Carlo particle 
   numbering scheme takes only into account mixtures of the third generation 
   squarks and the stau. For the other sfermions we will use the number of the 
   lefthanded field for the lighter mixed state and the one for the righthanded
   for the heavier. Below are the official pdg numbers from the Particle
   Data Group. In order not to produce arrays with some million entries in 
   the Fortran code for the masses and the widths we introduce our private 
   pdg numbering scheme which only extends not too far beyond 42. 
   Our private scheme then has the following pdf numbers (for the sparticles
   the subscripts $L/R$ and $1/2$ are taken synonymously): 

   \begin{center}
      \renewcommand{\arraystretch}{1.2}
       \begin{tabular}{|r|l|l|}\hline
         $d$                    & down-quark         &      1 \\\hline
         $u$                    & up-quark           &      2 \\\hline
         $s$                    & strange-quark      &      3 \\\hline
         $c$                    & charm-quark        &      4 \\\hline
         $b$                    & bottom-quark       &      5 \\\hline
         $t$                    & top-quark          &      6 \\\hline\hline
         $e^-$                  & electron           &     11 \\\hline
         $\nu_e$                & electron-neutrino  &     12 \\\hline
         $\mu^-$                & muon               &     13 \\\hline
         $\nu_\mu$              & muon-neutrino      &     14 \\\hline
         $\tau^-$               & tau                &     15 \\\hline
         $\nu_\tau$             & tau-neutrino       &     16 \\\hline\hline
         $g$                    & gluon              & (9) 21 \\\hline
         $\gamma$               & photon             &     22 \\\hline
         $Z^0$                  & Z-boson            &     23 \\\hline
         $W^+$                  & W-boson            &     24 \\\hline\hline
         $h^0$                  & light Higgs boson  &     25 \\\hline
         $H^0$                  & heavy Higgs boson  &     35 \\\hline
         $A^0$                  & pseudoscalar Higgs &     36 \\\hline
         $H^+$                  & charged Higgs      &     37 \\\hline\hline
         $\tilde{d}_L$          & down-squark 1      &     41 \\\hline 
         $\tilde{u}_L$          & up-squark 1        &     42 \\\hline
         $\tilde{s}_L$          & strange-squark 1   &     43 \\\hline
         $\tilde{c}_L$          & charm-squark 1     &     44 \\\hline
         $\tilde{b}_L$          & bottom-squark 1    &     45 \\\hline
         $\tilde{t}_L$          & top-squark 1       &     46 \\\hline
         $\tilde{d}_R$          & down-squark 2      &     47 \\\hline 
         $\tilde{u}_R$          & up-squark 2        &     48 \\\hline
         $\tilde{s}_R$          & strange-squark 2   &     49 \\\hline
         $\tilde{c}_R$          & charm-squark 2     &     50 \\\hline
         $\tilde{b}_R$          & bottom-squark 2    &     51 \\\hline
         $\tilde{t}_R$          & top-squark 2       &     52 \\\hline\hline
         $\tilde{e}_L$          & selectron 1        &     53 \\\hline
         $\tilde{\nu}_{e,L}$    & electron-sneutrino &     54 \\\hline
         $\tilde{\mu}_L$        & smuon 1            &     55 \\\hline
         $\tilde{\nu}_{\mu,L}$  & muon-sneutrino     &     56 \\\hline
         $\tilde{\tau}_L$       & stau 1             &     57 \\\hline
         $\tilde{\nu}_{\tau,L}$ & tau-sneutrino      &     58 \\\hline
         $\tilde{e}_R$          & selectron 2        &     59 \\\hline
         $\tilde{\mu}_R$        & smuon 2            &     61 \\\hline
         $\tilde{\tau}_R$       & stau 2             &     63 \\\hline\hline
         $\tilde{g}$            & gluino             &     64 \\\hline
         $\tilde{\chi}^0_1$     & neutralino 1       &     65 \\\hline
         $\tilde{\chi}^0_2$     & neutralino 2       &     66 \\\hline
         $\tilde{\chi}^0_3$     & neutralino 3       &     67 \\\hline
         $\tilde{\chi}^0_4$     & neutralino 4       &     68 \\\hline
         $\tilde{\chi}^0_4$     & neutralino 4       &     69 \\\hline
         $\tilde{\chi}^+_1$     & chargino 1         &     70 \\\hline
         $\tilde{\chi}^+_2$     & chargino 2         &     71 \\\hline\hline
         $a$                    & pseudoscalar       &     72 \\\hline
         $s$                    & scalar singlet     &     73 \\\hline
         $\tilde{G}$            & gravitino          &     -- \\\hline\hline 
     \end{tabular}
   \end{center}   *)

    let pdg_mw = function
      | L g when g > 0 -> 9 + 2*g
      | L g -> - 9 + 2*g
      | N g when g > 0 -> 10 + 2*g
      | N g -> - 10 + 2*g
      | U g when g > 0 -> 2*g
      | U g -> 2*g
      | D g when g > 0 -> - 1 + 2*g
      | D g -> 1 + 2*g
      | Gl -> 21
      | Ga -> 22 | Z -> 23
      | Wp -> 24 | Wm -> (-24)
      | H_Light -> 25 | H_Heavy -> 35 | A -> 36
      | Hp -> 37 | Hm -> (-37)
      | Sup (M1,g) when g > 0 -> 40 + 2*g
      | Sup (M1,g) -> - 40 + 2*g
      | Sup (M2,g) when g > 0 -> 46 + 2*g
      | Sup (M2,g) -> - 46 + 2*g
      | Sdown (M1,g) when g > 0 -> 39 + 2*g
      | Sdown (M1,g) -> - 39 + 2*g
      | Sdown (M2,g) when g > 0 -> 45 + 2*g
      | Sdown (M2,g) -> - 45 + 2*g           
      | Slepton (M1,g) when g > 0 -> 51 + 2*g
      | Slepton (M1,g) -> - 51 + 2*g
      | Slepton (M2,g) when g > 0 -> 57 + 2*g
      | Slepton (M2,g) -> - 57 + 2*g            
      | Sneutrino g when g > 0 ->  52 + 2*g
      | Sneutrino g -> - 52 + 2*g            
      | Gluino -> 64
      | Chargino C1 -> 70 | Chargino C1c -> (-70)
      | Chargino C2 -> 71 | Chargino C2c -> (-71)
      | Neutralino N1 -> 65 | Neutralino N2 -> 66
      | Neutralino N3 -> 67 | Neutralino N4 -> 68 
      | Neutralino N5 -> 69
      | P0 -> 72 | S0 -> 73

    let mass_symbol f =
      "mass(" ^ string_of_int (abs (pdg_mw f)) ^ ")"  

    let width_symbol f =
      "width(" ^ string_of_int (abs (pdg_mw f)) ^ ")"  

    let conj_symbol = function
      | false, str -> str
      | true, str -> str ^ "_c"

    let constant_symbol = function
      | Unit -> "unit" | Pi -> "PI"
      | Alpha_QED -> "alpha" | E -> "e" | G -> "g" | Vev -> "vev"
      | Sin2thw -> "sin2thw" | Eidelta -> "eidelta" | Mu -> "mu" | G_Z -> "gz"
      | Sin a -> "sin" ^ string_of_angle a | Cos a -> "cos" ^ string_of_angle a 
      | Sin2am2b -> "sin2am2b" | Cos2am2b -> "cos2am2b" | Sinamb -> "sinamb"
      | Sinapb -> "sinapb" | Cosamb -> "cosamb" | Cosapb -> "cosapb" 
      | Cos4be -> "cos4be" | Sin4be -> "sin4be" | Sin4al -> "sin4al" 
      | Sin2al -> "sin2al" | Cos2al -> "cos2al" | Sin2be -> "sin2be" 
      | Cos2be -> "cos2be" | Tana -> "tana" | Tanb -> "tanb"
      | Q_lepton -> "qlep" | Q_up -> "qup" | Q_down -> "qdwn"
      | Q_charg -> "qchar"
      | V_CKM (g1,g2) -> "vckm_" ^ string_of_int g1 ^ string_of_int g2
      | M_SF (f,g,m1,m2) -> "mix_" ^ string_of_sff f ^ string_of_int g
          ^ string_of_sfm m1 ^ string_of_sfm m2
      | AL g -> "al_" ^ string_of_int g 
      | AD g -> "ad_" ^ string_of_int g 
      | AU g -> "au_" ^ string_of_int g 
      | A_0 (n1,n2) -> "a0_" ^ string_of_neu n1 ^ string_of_neu n2
      | A_P (c1,c2) -> "ap_" ^ string_of_char c1 ^ string_of_char c2
      | V_0 (n1,n2) -> "v0_" ^ string_of_neu n1 ^ string_of_neu n2
      | V_P (c1,c2) -> "vp_" ^ string_of_char c1 ^ string_of_char c2
      | M_N (n1,n2) -> "mn_" ^ string_of_neu n1 ^ string_of_neu n2
      | M_U (c1,c2) -> "mu_" ^ string_of_char c1 ^ string_of_char c2
      | M_V (c1,c2) -> "mv_" ^ string_of_char c1 ^ string_of_char c2
      | L_NC (n,c) -> "lnc_" ^ string_of_neu n ^ string_of_char c
      | R_NC (n,c) -> "rnc_" ^ string_of_neu n ^ string_of_char c
      | L_CN (c,n) -> "lcn_" ^ string_of_char c ^ string_of_neu n
      | R_CN (c,n) -> "rcn_" ^ string_of_char c ^ string_of_neu n
      | L_NCH (n,c) -> "lnch_" ^ string_of_neu n ^ string_of_char c
      | R_NCH (n,c) -> "rnch_" ^ string_of_neu n ^ string_of_char c
      | L_CNG (c,n) -> "lcng_" ^ string_of_char c ^ string_of_neu n 
      | R_CNG (c,n) -> "rcng_" ^ string_of_char c ^ string_of_neu n 
      | S_NNA (n1,n2) -> "snna_" ^ string_of_neu n1 ^ string_of_neu n2
      | P_NNA (n1,n2) -> "pnna_" ^ string_of_neu n1 ^ string_of_neu n2
      | S_NNG (n1,n2) -> "snng_" ^ string_of_neu n1 ^ string_of_neu n2
      | P_NNG (n1,n2) -> "pnng_" ^ string_of_neu n1 ^ string_of_neu n2
      | S_NNH1 (n1,n2) -> "snnh1_" ^ string_of_neu n1 ^ string_of_neu n2
      | P_NNH1 (n1,n2) -> "pnnh1_" ^ string_of_neu n1 ^ string_of_neu n2
      | S_NNH2 (n1,n2) -> "snnh2_" ^ string_of_neu n1 ^ string_of_neu n2
      | P_NNH2 (n1,n2) -> "pnnh2_" ^ string_of_neu n1 ^ string_of_neu n2
      | G_NC_lepton -> "gnclep" | G_NC_neutrino -> "gncneu" 
      | G_NC_up -> "gncup" | G_NC_down -> "gncdwn"
      | G_CC -> "gcc"
      | G_CCQ (vc,g1,g2) -> conj_symbol (vc, "gccq_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | I_Q_W -> "iqw" | I_G_ZWW -> "igzww" 
      | G_WWWW -> "gw4" | G_ZZWW -> "gzzww"
      | G_PZWW -> "gpzww" | G_PPWW -> "gppww"   
      | G_GH 1 -> "ghaw" 
      | G_GH 2 -> "gh1az" | G_GH 3 -> "gh2az"
      | G_GH 4 -> "gh1ww" | G_GH 5 -> "gh2ww" 
      | G_GH 6 -> "ghh1w" | G_GH 7 -> "ghh2w" 
      | G_GH 8 -> "gh1zz" | G_GH 9 -> "gh2zz" 
      | G_GH 10 -> "ghhz" | G_GH 11 -> "ghhp"            
      | G_GHGo n -> "g_hgh(" ^ string_of_int n ^ ")"  
      | G_GH4 1 -> "gaazz" | G_GH4 2 -> "gh1h1zz" | G_GH4 3 -> "gh2h2zz"
      | G_GH4 4 -> "ghphmzz" | G_GH4 5 -> "ghphmpp" | G_GH4 6 -> "ghphmpz"
      | G_GH4 7 -> "ghh1wz" | G_GH4 8 -> "ghh2wz"
      | G_GH4 9 -> "ghh1wp" | G_GH4 10 -> "ghh2wp" 
      | G_GH4 11 -> "gaaww" | G_GH4 12 -> "gh1h1ww" | G_GH4 13 -> "gh2h2ww" 
      | G_GH4 14 -> "ghhww" | G_GH4 15 -> "ghawz" | G_GH4 16 -> "ghawp" 
      | G_CICIH1 (n1,n2) -> "gcicih1_" ^ string_of_neu n1 ^ "_" 
          ^ string_of_neu n2
      | G_CICIH2 (n1,n2) -> "gcicih2_" ^ string_of_neu n1 ^ "_" 
          ^ string_of_neu n2 
      | G_CICIA (n1,n2) -> "gcicia_" ^ string_of_neu n1 ^ "_" 
          ^ string_of_neu n2
      | G_CICIG (n1,n2) -> "gcicig_" ^ string_of_neu n1 ^ "_" 
          ^ string_of_neu n2 
      | G_H3 n -> "gh3_" ^ string_of_int n
      | G_H4 n -> "gh4_" ^ string_of_int n 
      | G_HGo3 n -> "ghg3_" ^ string_of_int n 
      | G_HGo4 n -> "ghg4_" ^ string_of_int n 
      | G_GG4 n -> "ggg4_" ^ string_of_int n
      | G_strong -> "gs" | G_SS -> "gs**2" 
      | Gs -> "gs"
      | I_G_S -> "igs"
      | G_S_Sqrt -> "gssq"
      | G_NWC (n,c) -> "gnwc_" ^ string_of_neu n ^ "_" ^ string_of_char c 
      | G_CWN (c,n) -> "gcwn_" ^ string_of_char c ^ "_" ^ string_of_neu n 
      | G_CH1C (c1,c2) -> "gch1c_" ^ string_of_char c1 ^ "_" ^ string_of_char c2
      | G_CH2C (c1,c2) -> "gch2c_" ^ string_of_char c1 ^ "_" ^ string_of_char c2
      | G_CAC (c1,c2) -> "gcac_" ^ string_of_char c1 ^ "_" ^ string_of_char c2
      | G_CGC (c1,c2) -> "gcgc_" ^ string_of_char c1 ^ "_" ^ string_of_char c2
      | G_YUK (i,g) -> "g_yuk" ^ string_of_int i ^ "_" ^ string_of_int g
      | G_NZN (n1,n2) -> "gnzn_" ^ string_of_neu n1 ^ "_" ^ string_of_neu n2 
      | G_CZC (c1,c2) -> "gczc_" ^ string_of_char c1 ^ "_" ^ string_of_char 
          c2 
      | DUM 1 -> "dummy1" | DUM 5 -> "dummy5" 
      | G_YUK_1 (n,m) -> "g_yuk1_" ^ string_of_int n ^ "_" ^ string_of_int m 
      | G_YUK_2 (n,m) -> "g_yuk2_" ^ string_of_int n ^ "_" ^ string_of_int m 
      | G_YUK_3 (n,m) -> "g_yuk3_" ^ string_of_int n ^ "_" ^ string_of_int m 
      | G_YUK_4 (n,m) -> "g_yuk4_" ^ string_of_int n ^ "_" ^ string_of_int m 
      | G_YUK_C (vc,g,c,sf,m) -> conj_symbol (vc, "g_yuk_ch" ^ string_of_char c  
          ^ "_" ^ string_of_sff sf ^ string_of_sfm m ^ "_" ^ string_of_int g ) 
      | G_YUK_N (vc,g,n,sf,m) -> conj_symbol (vc, "g_yuk_n" ^ string_of_neu n
          ^ "_" ^ string_of_sff sf ^ string_of_sfm m ^ "_" ^ string_of_int g )
      | G_YUK_G (vc,g,sf,m) -> conj_symbol (vc, "g_yuk_g" ^ string_of_sff sf 
          ^ string_of_sfm m ^ "_" ^ string_of_int g)
      | G_YUK_Q (vc,g1,g2,c,sf,m) -> conj_symbol (vc, "g_yuk_ch" ^ string_of_char c  
          ^ "_" ^ string_of_sff sf ^ string_of_sfm m ^ "_" ^ string_of_int g1 
          ^ "_" ^ string_of_int g2) 
      | G_NHC (n,c) -> "g_nhc_" ^ string_of_neu n ^ "_" ^ string_of_char c 
      | G_CHN (c,n) -> "g_chn_" ^ string_of_neu n ^ "_" ^ string_of_char c
      | G_NGC (n,c) -> "g_ngc_" ^ string_of_neu n ^ string_of_char c
      | G_CGN (c,n) -> "g_cgn_" ^ string_of_char c ^ string_of_neu n
      | SUM_1 -> "sum1"
      | G_SLSNW (vc,g,m) -> conj_symbol (vc, "gsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int g ^ "snw") 
      | G_ZSF (f,g,m1,m2) -> "g" ^ string_of_sff f ^ string_of_sfm m1 ^ "z" 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_WWSFSF (f,g,m1,m2) -> "gww" ^ string_of_sff f ^ string_of_sfm m1   
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_WPSLSN (vc,g,m) -> conj_symbol (vc, "gpwsl" ^ string_of_sfm m 
          ^ "sn_" ^ string_of_int g)
      | G_WZSLSN (vc,g,m) -> conj_symbol (vc, "gwzsl" ^ string_of_sfm m 
          ^ "sn_" ^ string_of_int g)
      | G_H1SFSF (f,g,m1,m2) -> "gh1" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g    
      | G_H2SFSF (f,g,m1,m2) -> "gh2" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g  
      | G_ASFSF (f,g,m1,m2) -> "ga" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g
      | G_HSNSL (vc,g,m) -> conj_symbol (vc, "ghsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int g)
      | G_GoSFSF (f,g,m1,m2) -> "ggo" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_GoSNSL (vc,g,m) -> conj_symbol (vc, "ggosnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int g) 
      | G_HSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "ghsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_GSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "ggsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_WPSUSD (vc,m1,m2,n,m) -> conj_symbol (vc, "gpwpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int n ^ "_" 
          ^ string_of_int m)
      | G_WZSUSD (vc,m1,m2,n,m) -> conj_symbol (vc, "gzwpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int n ^ "_" 
          ^ string_of_int m)
      | G_SWS (vc,g1,g2,m1,m2) -> conj_symbol (vc, "gs" ^ string_of_sfm m1 ^ "ws" 
          ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" ^ string_of_int g2)
      | G_GlGlSQSQ -> "gglglsqsq" 
      | G_PPSFSF f -> "gpp" ^ string_of_sff f ^ string_of_sff f 
      | G_ZZSFSF (f,g,m1,m2) -> "gzz" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_ZPSFSF (f,g,m1,m2) -> "gzp" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_GlPSQSQ -> "gglpsqsq" 
      | G_GlZSFSF (f,g,m1,m2) -> "ggl" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int g 
      | G_GlWSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gglwsu" 
          ^ string_of_sfm m1 ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 
          ^ "_" ^ string_of_int g2)
      | G_GHGo4 1 -> "gzzg0g0" | G_GHGo4 2 -> "gzzgpgm" 
      | G_GHGo4 3 -> "gppgpgm" | G_GHGo4 4 -> "gzpgpgm" 
      | G_GHGo4 5 -> "gwwgpgm" | G_GHGo4 6 -> "gwwg0g0"
      | G_GHGo4 7 -> "gwzg0g" | G_GHGo4 8 -> "gwzg0g" 
      | G_GHGo4 9 -> "gwzh1g" | G_GHGo4 10 -> "gwzh2g"
      | G_GHGo4 11 -> "gwph1g" | G_GHGo4 12 -> "gwph2g"   
      | G_HSF31 (h,g,m1,m2,f1,f2) -> "g_" ^ string_of_higgs h ^ 
          string_of_int g ^ string_of_sfm m1 ^ string_of_sfm m2 ^
          string_of_sff f1 ^ string_of_sff f2
      | G_HSF32 (h,g1,g2,m1,m2,f1,f2) -> "g_" ^ string_of_higgs h ^ 
          string_of_int g1 ^ "_" ^ string_of_int g2 ^ string_of_sfm m1 ^ 
          string_of_sfm m2 ^ string_of_sff f1 ^ string_of_sff f2
      | G_HSF41 (h,g,m1,m2,f1,f2) -> "g_" ^ string_of_higgs h ^ 
          string_of_int g ^ string_of_sfm m1 ^ string_of_sfm m2 ^
          string_of_sff f1 ^ string_of_sff f2
      | G_HSF42 (h,g1,g2,m1,m2,f1,f2) -> "g_" ^ string_of_higgs h ^ 
          string_of_int g1 ^ "_" ^ string_of_int g2 ^ string_of_sfm m1 ^ 
          string_of_sfm m2 ^ string_of_sff f1 ^ string_of_sff f2
      | G_H1H1SFSF (f,m1,m2,n) -> "gh1h1" ^ string_of_sff f ^ string_of_sfm 
          m1 ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n      
      | G_H1H2SFSF (f,m1,m2,n) -> "gh1h2" ^ string_of_sff f ^ string_of_sfm 
          m1 ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n   
      | G_H2H2SFSF (f,m1,m2,n) -> "gh2h2" ^ string_of_sff f ^ string_of_sfm 
          m1 ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n  
      | G_HHSFSF (f,m1,m2,n) -> "ghh" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n  
      | G_AASFSF (f,m1,m2,n) -> "gaa" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n 
      | G_HH1SUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "ghh1su" 
          ^ string_of_sfm m1 ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 
          ^ "_" ^ string_of_int g2)
      | G_HH2SUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "ghh2su" 
          ^ string_of_sfm m1 ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 
          ^ "_" ^ string_of_int g2)
      | G_HASUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "ghasu" 
          ^ string_of_sfm m1 ^ "sd" ^ string_of_sfm m2 ^ "_" 
          ^ string_of_int g1 ^ "_" ^ string_of_int g2 ^ "_c")
      | G_HH1SLSN (vc,m,g) -> conj_symbol (vc, "ghh1sl" ^ string_of_sfm m 
                                           ^ "sn_" ^ string_of_int g)
      | G_HH2SLSN (vc,m,g) -> conj_symbol (vc, "ghh2sl" ^ string_of_sfm m 
                                           ^ "sn_" ^ string_of_int g)
      | G_HASLSN (vc,m,g) -> conj_symbol (vc, "ghasl" ^ string_of_sfm m  
                                           ^ "sn_" ^ string_of_int g)
      | G_AG0SFSF (f,m1,m2,n) -> "gag0" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n  
      | G_HGSFSF (f,m1,m2,n) -> "ghg" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m1 ^ "_" ^ string_of_int n  
      | G_GGSFSF (f,m1,m2,n) -> "ggg" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n  
      | G_G0G0SFSF (f,m1,m2,n) -> "gg0g0" ^ string_of_sff f ^ string_of_sfm m1 
          ^ string_of_sff f ^ string_of_sfm m2 ^ "_" ^ string_of_int n  
      | G_HGSNSL (vc,m,n) -> conj_symbol (vc, "ghgsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int n)
      | G_H1GSNSL (vc,m,n) -> conj_symbol (vc, "gh1gsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int n)
      | G_H2GSNSL (vc,m,n) -> conj_symbol (vc, "gh2gsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int n) 
      | G_AGSNSL (vc,m,n) -> conj_symbol (vc, "gagsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int n) 
      | G_GGSNSL (vc,m,n) -> conj_symbol (vc, "gggsnsl" ^ string_of_sfm m ^ "_" 
          ^ string_of_int n) 
      | G_HGSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gghpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_H1GSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gh1gpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_H2GSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gh2gpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_AGSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gagpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_GGSUSD (vc,m1,m2,g1,g2) -> conj_symbol (vc, "gggpsu" ^ string_of_sfm m1 
          ^ "sd" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 ^ "_" 
          ^ string_of_int g2)
      | G_SN4 (g1,g2) -> "gsn4_" ^ string_of_int g1 ^ "_" ^ string_of_int g2 
      | G_SN2SL2_1 (m1,m2,g1,g2) -> "gsl_" ^ string_of_int g1 ^ "_sl_" 
          ^ string_of_int g1 ^ "_sl" ^ string_of_sfm m1 ^ "_" ^ string_of_int g2 
          ^ "_sl" ^ string_of_sfm m2 ^ "_" ^ string_of_int g2 
      | G_SN2SL2_2 (m1,m2,g1,g2) -> "gsl_" ^ string_of_int g1 ^ "_sl_" 
          ^ string_of_int g2 ^ "_sl" ^ string_of_sfm m1 ^ "_" ^ string_of_int g1 
          ^ "_sl" ^ string_of_sfm m2 ^ "_" ^ string_of_int g2 ^ "_mix"
      | G_SF4 (f1,f2,m1,m2,m3,m4,g1,g2) -> "gsf" ^ string_of_sff f1 ^ 
          string_of_sff f2 ^ string_of_sfm m1 ^ string_of_sfm m2 ^ 
          string_of_sfm m3 ^ string_of_sfm m4 ^ string_of_int g1 ^
          string_of_int g2 
      | G_SF4_3 (f1,f2,m1,m2,m3,m4,g1,g2,g3) -> "gsf" ^ string_of_sff f1 ^ 
          string_of_sff f2 ^ string_of_sfm m1 ^ string_of_sfm m2 ^ 
          string_of_sfm m3 ^ string_of_sfm m4 ^ string_of_int g1 ^
          string_of_int g2 ^ "_" ^ string_of_int g3 
      | G_SF4_4 (f1,f2,m1,m2,m3,m4,g1,g2,g3,g4) -> "gsf" ^ string_of_sff f1 ^ 
          string_of_sff f2 ^ string_of_sfm m1 ^ string_of_sfm m2 ^ 
          string_of_sfm m3 ^ string_of_sfm m4 ^ string_of_int g1 ^ "_" ^ 
          string_of_int g2 ^ string_of_int g3 ^ "_" ^ string_of_int g4
      | G_SL4 (m1,m2,m3,m4,g) -> "gsl" ^ string_of_sfm m1 ^ "_" 
          ^ "sl" ^ string_of_sfm m2 ^ "_" ^ "sl" ^ string_of_sfm m3 ^ "_" 
          ^ "sl" ^ string_of_sfm m4 ^ "_" ^ string_of_int g 
      | G_SL4_2 (m1,m2,m3,m4,g1,g2) -> "gsl" ^ string_of_sfm m1 ^ "_" 
          ^ "sl" ^ string_of_sfm m2 ^ "_" ^ "sl" ^ string_of_sfm m3 ^ "_" 
          ^ "sl" ^ string_of_sfm m4 ^ "_" ^ string_of_int g1 ^ "_" ^
          string_of_int g2
      | G_SN2SQ2 (f,m1,m2,g1,g2) -> "gsn_" ^ string_of_int g1 ^ "_sn_" 
          ^ string_of_int g1 ^ "_" ^ string_of_sff f ^ string_of_sfm m1 ^ "_" 
          ^ string_of_int g2 ^ "_" ^ string_of_sff f ^ string_of_sfm m2 ^ "_" 
          ^ string_of_int g2
      | G_SL2SQ2 (f,m1,m2,m3,m4,g1,g2) -> "gsl" ^ string_of_sfm m1 ^ "_" 
          ^ string_of_int g1 ^ "_sl" ^ string_of_sfm m2 ^ "_" ^ string_of_int g1 
          ^ "_" ^ string_of_sff f ^ string_of_sfm m3 ^ "_" ^ string_of_int g2 
          ^ "_" ^ string_of_sff f ^ string_of_sfm m4 ^ "_" ^ string_of_int g2 
      | G_SUSDSNSL (vc,m1,m2,m3,g1,g2,g3) -> conj_symbol (vc, "gsl" 
          ^ string_of_sfm m3 ^ "_" ^ string_of_int g3 ^ "_sn_" ^ string_of_int g3 
          ^ "_su" ^ string_of_sfm m1 ^ "_" ^ string_of_int g1 ^ "_sd" 
          ^ string_of_sfm m2 ^ "_" ^ string_of_int g2) 
      | G_SU4 (m1,m2,m3,m4,g) -> "gsu" ^ string_of_sfm m1 ^ "_" 
          ^ "_su" ^ string_of_sfm m2 ^ "_" ^ "_su" ^ string_of_sfm m3 ^ "_" ^ 
          "_su" ^ string_of_sfm m4 ^ "_" ^ string_of_int g 
      | G_SU4_2 (m1,m2,m3,m4,g1,g2) -> "gsu" ^ string_of_sfm m1 ^ "_" 
          ^ "_su" ^ string_of_sfm m2 ^ "_" ^ "_su" ^ string_of_sfm m3 ^ "_" ^ 
          "_su" ^ string_of_sfm m4 ^ "_" ^ string_of_int g1 ^ "_" ^ 
          string_of_int g2
      | G_SD4 (m1,m2,m3,m4,g) -> "gsd" ^ string_of_sfm m1 ^ "_" 
          ^ "_sd" ^ string_of_sfm m2 ^ "_" ^ "_sd" ^ string_of_sfm m3 ^ "_" 
          ^ "_sd" ^ string_of_sfm m4 ^ "_" ^ string_of_int g 
      | G_SD4_2 (m1,m2,m3,m4,g1,g2) -> "gsd" ^ string_of_sfm m1 ^ "_" 
          ^ "_sd" ^ string_of_sfm m2 ^ "_" ^ "_sd" ^ string_of_sfm m3 ^ "_" 
          ^ "_sd" ^ string_of_sfm m4 ^ "_" ^ string_of_int g1 ^ "_" ^ 
          string_of_int g2
      | G_SU2SD2 (m1,m2,m3,m4,g1,g2,g3,g4) -> "gsu" ^ string_of_sfm m1 
          ^ "_" ^ string_of_int g1 ^ "_su" ^ string_of_sfm m2 ^ "_" 
          ^ string_of_int g2 ^ "_sd" ^ string_of_sfm m3 ^ "_" ^ string_of_int g3 
          ^ "_sd" ^ string_of_sfm m4 ^ "_" ^ string_of_int g4   
      | M f -> "mass" ^ flavor_symbol f
      | W f -> "width" ^ flavor_symbol f 
      | _ -> "Panic: not available" 
      
  end

(*i
 *  Local Variables:
 *  mode:caml
 *  indent-tabs-mode:nil
 *  page-delimiter:"^(\\* .*\n"
 *  End:
i*)
